"""This module provides some known ANSI escape codes which work with terminals

All numbers used in this file are values usable in an ANSI escape sequence
    See http://en.wikipedia.org/wiki/ANSI_escape_code for values
"""


def escape(string):
    return '\033%s' % string


def escape_sequence(string):
    return escape('[%sm' % string)


def bold():
    return escape_sequence('1')


def no_bold():
    return escape_sequence('22')


def no_colour():
    return escape_sequence('2') + escape_sequence('0')


def no_prompt():
    return prompt(no_colour())


def _colour_16(ground, i):
    if i > 7:
        prefix = bold()
        i = i - 8
    else:
        prefix = ''
    return '%s%s' % (prefix, escape('[%d%dm' % (ground, i)))


def _colour_256(ground, i):
    return escape_sequence('%s;5;%d') % (ground, i)


def _background_16(i):
    return _colour_16(4, i)


def _background_256(i):
    return _colour_256(48, i)


def _foreground_16(i):
    return _colour_16(3, i)


def _foreground_256(i):
    return _colour_256(38, i)


def _small_colour_number(i):
    return i < 16


def foreground(i):
    return _small_colour_number(i) and _foreground_16(i) or _foreground_256(i)


def background(i):
    return _small_colour_number(i) and _background_16(i) or _background_256(i)


def prompt(string):
    return '\001%s\002' % string


def foreground_string(text, i):
    return '%s%s%s' % (foreground(i), text, no_colour())


def background_string(text, i):
    return '%s%s%s' % (background(i), text, no_colour())


def grounds_string(text, background_colour, foreground_colour):
    return '%s%s%s%s' % (
        background(background_colour),
        foreground(foreground_colour),
        text,
        no_colour())


def prompt_string(text, i):
    string = foreground(i)
    return '%s%s%s' % (prompt(string), text, no_prompt())



"""This module handles lists"""


def de_duplicate(items):
    """Remove any duplicate item, preserving order

    >>> de_duplicate([1, 2, 1, 2])
    [1, 2]
    """
    result = []
    for item in items:
        if item not in result:
            result.append(item)
    return result



"""A collection of extended path classes and methods

The classes all inherit from the original path.path
"""
import os
from fnmatch import fnmatch


class PathError(Exception):
    """Something went wrong with a path"""
    prefix = 'Path Error'

    def __init__(self, message):
        Exception.__init__(self, message)


class PathAssertions(object):
    """Assertions that can be made about paths"""
    # pylint: disable=no-member

    def assert_exists(self):
        """Raise a PathError if this path does not exist on disk"""
        if not self.exists():
            raise PathError('%s does not exist' % self)
        return self

    def assert_isdir(self):
        """Raise a PathError if this path is not a directory on disk"""
        if not self.isdir():
            raise PathError('%s is not a directory' % self)
        return self

    def assert_isfile(self):
        """Raise a PathError if this path is not a file on disk"""
        if not self.isfile():
            raise PathError('%s is not a file' % self)
        return self


from path import Path as PPath  # pylint: disable=wrong-import-position


class DotPath(PPath):
    """Some additions to the classic path class"""
    # pylint: disable=abstract-method

    def __repr__(self):
        return '<%s %s>' % (
            self.__class__.__name__,
            repr(str(self)))

    # The / operator joins paths.
    def __div__(self, child):
        """Join two path components, adding a separator character if needed.

        If the result is a file return self.__file_class__(result)

        >>> p = DotPath('/home/guido')
        >>> p.__div__('fred') == p / 'fred' == p.joinpath('fred')
        True
        """
        if child:
            result = os.path.join(self, child)
        else:
            result = str(self)  # pylint: disable=redefined-variable-type
        return self.as_existing_file(result)

    __truediv__ = __div__

    def __cmp__(self, other):
        return cmp(str(self), str(other))

    def _next_class(self, string):
        return self.as_existing_file(string)

    def as_existing_file(self, filepath):
        """Return the file class for existing files only"""
        if os.path.isfile(filepath) and hasattr(self, '__file_class__'):
            return self.__file_class__(filepath)  # pylint: disable=no-member
        return self.__class__(filepath)

    def parent_directory(self):
        if str(self) == '/':
            return None
        return self.parent

    def directory(self):
        """Return a path to the path's directory"""
        return self.parent

    def dirnames(self):
        """Split the dirname into individual directory names

        An absolute path starts with an empty string, a relative path does not

        >>> DotPath(u'/path/to/module.py').dirnames() == [u'', u'path', u'to']
        True
        >>> DotPath(u'path/to/module.py').dirnames() == [u'path', u'to']
        True
        """
        return self.dirname().split(os.path.sep)

    def dirpaths(self):
        """Split the dirname into individual directory names

        An absolute path starts with an empty string, a relative path does not

        >>> p = DotPath(u'/path/to/x.py')
        >>> p.paths == p.dirpaths()
        True
        """
        parts = self.parts()
        result = [DotPath(parts[0] or '/')]
        for name in parts[1:]:
            result.append(result[-1] / name)
        return result

    def parts(self):
        """Split the path into parts like Pathlib

        >>> expected = ['/', 'path', 'to', 'there']
        >>> assert DotPath('/path/to/there').parts() == expected
        """
        parts = self.split(os.path.sep)
        parts[0] = parts[0] and parts[0] or '/'
        return parts

    def directories(self):
        """Split the dirname into individual directory names

        No empty parts are included

        >>> DotPath(u'path/to/module.py').directories() == [u'path', u'to']
        True
        >>> DotPath(u'/path/to/module.py').directories() == [u'path', u'to']
        True
        """
        return [d for d in self.dirnames() if d]

    parents = property(
        dirnames, None, None,
        """ This path's parent directories, as a list of strings.

        >>> DotPath(u'/path/to/module.py').parents == [u'', u'path', u'to']
        True
        """)

    paths = property(
        dirpaths, None, None,
        """ This path's parent directories, as a sequence of paths.

        >>> paths = DotPath('/usr/bin/vim').paths
        >>> paths[-1].isfile()  # vim might be a link
        True
        >>> paths[-2] == paths[-1].parent
        True
        >>> paths[-3] == paths[-2].parent
        True
        >>> paths[-4] == paths[-3].parent
        True
        >>> paths[-4] == paths[0]
        True
        """)

    def short_relative_path_to(self, destination):
        """The shorter of either the absolute path of the destination,
            or the relative path to it

        >>> print(DotPath('/home/guido/bin').short_relative_path_to(
        ...     '/home/guido/build/python.tar'))
        ../build/python.tar
        >>> print(DotPath('/home/guido/bin').short_relative_path_to(
        ...     '/mnt/guido/build/python.tar'))
        /mnt/guido/build/python.tar
        """
        relative = self.relpathto(destination)
        absolute = self.__class__(destination).abspath()
        if len(relative) < len(absolute):
            return relative
        return absolute

    def short_relative_path_to_here(self):
        """A short path relative to current working directory"""
        return self.short_relative_path_to(os.getcwd())

    def short_relative_path_from_here(self):
        """A short path relative to self to the current working directory"""
        return self.__class__(os.getcwd()).short_relative_path_to(self)

    def walk_some_dirs(self, levels=-1, pattern=None):
        """ D.walkdirs() -> iterator over subdirs, recursively.

        With the optional 'pattern' argument, this yields only
        directories whose names match the given pattern.  For
        example, mydir.walkdirs('*test') yields only directories
        with names ending in 'test'.
        """
        if not levels:
            yield self
            raise StopIteration
        levels -= 1
        for child in self.dirs():
            if pattern is None or child.fnmatch(pattern):
                yield child
            if levels:
                for subsubdir in child.walk_some_dirs(levels, pattern):
                    yield subsubdir

    def fnmatch_basename(self, glob):
        if glob.startswith(os.path.sep):
            glob = glob.lstrip(os.path.sep)
        string = self.basename()
        if fnmatch(string, glob):
            return self
        return None

    def fnmatch_directory(self, glob):
        if glob.startswith(os.path.sep) or glob.endswith(os.path.sep):
            glob = glob.strip(os.path.sep)
        if self.isdir():
            string = self.basename()
        else:
            string = self.parent.basename()
        if fnmatch(string, glob):
            return self
        return None

    def fnmatch_directories(self, glob):
        if glob.startswith(os.path.sep) or glob.endswith(os.path.sep):
            glob = glob.strip(os.path.sep)
        else:
            return
        strings = reversed(self.directory().splitall()[1:])
        for string in strings:
            if fnmatch(string, glob):
                return self
        return None

    def fnmatch_part(self, glob):
        if self.fnmatch(glob):
            return self
        if self.fnmatch_basename(glob):
            return self
        if self.fnmatch_directory(glob):
            return self
        if self.fnmatch_directories(glob):
            return self
        return None

    def __get_owner_not_implemented(self):
        pass

    def expandall(self):
        try:
            return self.expand().realpath().abspath()
        except AttributeError:
            return self.__class__(
                os.path.abspath(os.path.realpath(
                    os.path.expanduser(os.path.expandvars(str(self)))
                ))
            )

    def same_path(self, other):
        return self.expandall() == other.expandall()

    @property
    def hidden(self):
        """A 'hidden file' has a name starting with a '.'"""
        s = str(self.basename())
        return s and s[0] == '.'


def ext_language(ext, exts=None):
    """Language of the extension in those extensions

    If exts is supplied, then restrict recognition to those exts only
    If exts is not supplied, then use all known extensions

    >>> ext_language('.py') == 'python'
    True
    """
    languages = {
        '.py': 'python',
        '.py2': 'python2',
        '.py3': 'python3',
        '.sh': 'bash',
        '.bash': 'bash',
        '.pl': 'perl',
        '.js': 'javascript',
        '.txt': 'english',
    }
    ext_languages = {_: languages[_] for _ in exts} if exts else languages
    return ext_languages.get(ext)


def find_language(script, exts=None):
    """Determine the script's language  extension

    >>> this_script = __file__.rstrip('c')
    >>> find_language(makepath(this_script)) == 'python'
    True

    If exts are given they restrict which extensions are allowed
    >>> find_language(makepath(this_script), ('.sh', '.txt')) is None
    True

    If there is no extension, but shebang is present, then use that
    (Expecting to find "#!" in ~/.bashrc for this test,
        but ~/.bashrc might not exist - then there's no language)

    >>> bashrc = home() / '.bashrc'
    >>> find_language(bashrc) in ('bash', None)
    True
    """
    if not script.isfile():
        return None
    if script.ext:
        return ext_language(script.ext, exts)
    shebang = script.shebang()
    return shebang and str(shebang.name) or None


del PPath


class FilePath(DotPath, PathAssertions):
    """A path to a known file"""

    def __div__(self, child):
        raise PathError('%r has no children' % self)

    __truediv__ = __div__

    def __iter__(self):
        for line in self.stripped_lines():
            yield line

    def try_remove(self):
        """Try to remove the file"""
        self.remove()

    def stripped_lines(self):
        """A list of all lines without trailing whitespace

        If lines can not be read (e.g. no such file) then an empty list
        """
        try:
            return [l.rstrip() for l in self.lines(retain=False)]
        except IOError:
            return []

    def stripped_whole_lines(self):
        """A list of all lines without trailing whitespace or blank lines"""
        return [l for l in self.stripped_lines() if l]

    def non_comment_lines(self):
        """A list of all non-empty, non-comment lines"""
        return [l
                for l in self.stripped_whole_lines()
                if not l.startswith('#')]

    def has_line(self, string):
        for line in self:
            if string == line:
                return True
        return False

    def any_line_has(self, string):
        for line in self:
            if string in line:
                return True
        return False

    def split_all_ext(self):
        copy = self[:]
        filename, ext = os.path.splitext(copy)
        if ext == '.gz':
            filename, ext = os.path.splitext(filename)
            ext = '%s.gz' % ext
        return self.__class__(filename), ext

    def as_python(self):
        """The path to the file with a .py extension

        >>> FilePath('/path/to/fred.txt').as_python()
        <FilePath '/path/to/fred.py'>
        """
        return self.extend_by('.py')

    def extend_by(self, extension):
        """The path to the file changed to use the given extension

        >>> FilePath('/path/to/fred').extend_by('.txt')
        <FilePath '/path/to/fred.txt'>
        >>> FilePath('/path/to/fred.txt').extend_by('..tmp')
        <FilePath '/path/to/fred.tmp'>
        >>> FilePath('/path/to/fred.txt').extend_by('fred')
        <FilePath '/path/to/fred.fred'>
        """
        copy = self[:]
        filename, _ = os.path.splitext(copy)
        return self.__class__('%s.%s' % (filename, extension.lstrip('.')))

    def make_read_only(self):
        """chmod the file permissions to -r--r--r--"""
        self.chmod(ChmodValues.readonly_file)

    def cd(self):  # pylint: disable=invalid-name
        """Change program's current directory to self"""
        return cd(self.parent)

    def dirname(self):
        return DirectPath(os.path.dirname(self))
    parent = property(dirname)

    def shebang(self):
        """The  #! entry from the first line of the file

        If no shebang is present, return an empty string
        """
        try:
            try:
                first_line = self.lines(retain=False)[0]
            except (OSError, UnicodeDecodeError):
                return ''
            if first_line.startswith('#!'):
                rest_of_line = first_line[2:].strip()
                parts = rest_of_line.split(' ')
                interpreter = parts[0]
                return makepath(interpreter)
        except IndexError:
            pass
        return ''

    def mv(self, destination):  # pylint: disable=invalid-name
        return self.move(destination)

    def make_file_exist(self):
        """Make sure the parent directory exists, then touch the file"""
        self.parent.make_directory_exist()
        self.parent.touch_file(self.name)
        return self

    @property
    def language(self):
        """The language of this file"""
        try:
            return self._language
        except AttributeError:
            self._language = find_language(self, getattr(self, 'exts', None))
        return self._language

    @language.setter
    def language(self, value):
        self._langauge = value

    def choose_language(self, exts):
        self._exts = exts
        self._language = ext_language(self.ext, exts)


class DirectPath(DotPath, PathAssertions):
    """A path which knows it might be a directory

    And that files are in directories
    """

    __file_class__ = FilePath

    def __iter__(self):
        for a_path in DotPath.listdir(self):
            yield a_path

    def directory(self):
        """Return a path to a directory.

        Either the path itself (if it is a directory), or its parent)
        """
        if self.isdir():
            return self
        return self.parent

    def try_remove(self):
        """Try to remove the path

        If it is a directory, try recursive removal of contents too
        """
        if self.islink():
            self.unlink()
        elif self.isfile():
            self.remove()
        elif self.isdir():
            self.empty_directory()
            if self.isdir():
                self.rmdir()
        else:
            return False
        return True

    def empty_directory(self):
        """Remove all contents of a directory

        Including any sub-directories and their contents"""
        for child in self.walkfiles():
            child.remove()
        for child in reversed([d for d in self.walkdirs()]):
            if child == self or not child.isdir():
                continue
            child.rmdir()

    def cd(self):  # pylint: disable=invalid-name
        """Change program's current directory to self"""
        return cd(self)

    def listdir(self, pattern=None):
        return [self.as_existing_file(_)
                for _ in DotPath.listdir(self, pattern)]

    def list_dirs(self, pattern=None):
        return self.list_dirs_files(pattern)[0]

    def list_files(self, pattern=None):
        return self.list_dirs_files(pattern)[1]

    def list_dirsfiles(self, pattern=None):
        dirs, others = self.list_dirs_files(pattern)
        return dirs + others

    def list_dirs_files(self, pattern=None):
        items = self.listdir(pattern)
        dirs = [_ for _ in items if _.isdir()]
        others = [_ for _ in items if not _.isdir()]
        return dirs, others

    def make_directory_exist(self):
        if self.isdir():
            return False
        if os.path.exists(self):
            raise PathError('%s exists but is not a directory' % self)
        self.makedirs()

    def make_file_exist(self, filename=None):
        """Make the directory exist, then touch the file

        If the filename is None, then use self.name as filename
        """
        if filename is None:
            path_to_file = FilePath(self)
            path_to_file.make_file_exist()
            return path_to_file
        else:
            self.make_directory_exist()
            path_to_file = self.touch_file(filename)
            return FilePath(path_to_file)

    def make_read_only(self):
        """chmod the directory permissions to -r-xr-xr-x"""
        self.chmod(ChmodValues.readonly_directory)

    def touch_file(self, filename):
        """Touch a file in the directory"""
        path_to_file = self.__file_class__(os.path.join(self, filename))
        path_to_file.touch()
        return path_to_file

    def existing_sub_paths(self, sub_paths):
        """Those in the given list of sub_paths which do exist"""
        paths_to_subs = [self / _ for _ in sub_paths]
        return [_ for _ in paths_to_subs if _.exists()]

    def clear_directory(self):
        """Make sure the directory exists and is empty"""
        self.make_directory_exist()
        self.empty_directory()

    # pylint: disable=arguments-differ
    def walkfiles(self, pattern=None, errors='strict', ignores=None):
        ignored = ignore_globs(ignores)
        for path_to_file in super(DirectPath, self).walkfiles(pattern, errors):
            if not ignored(path_to_file):
                yield path_to_file

    def listfiles(self, pattern=None, ignores=None):
        ignored = ignore_globs(ignores)
        return [_ for _ in self.listdir(pattern) if _.isfile() and not ignored(_)]



def ignore_globs(ignores):

    def ignored(a_path):
        if not ignores:
            return False
        for ignore in ignores:
            if a_path.fnmatch_part(ignore):
                return True
        return False

    return ignored


class ChmodValues(object):
    # pylint: disable=too-few-public-methods
    readonly_file = 0o444
    readonly_directory = 0o555


def makepath(s, as_file=False):
    """Make a path from a string

    Expand out any variables, home squiggles, and normalise it
    See also http://stackoverflow.com/questions/26403972
    """
    if s is None:
        return None
    result = FilePath(s) if (os.path.isfile(s) or as_file) else DirectPath(s)
    return result.expandall()

path = makepath  # pylint: disable=invalid-name


def cd(path_to):  # pylint: disable=invalid-name
    """cd to the given path

    If the path is a file, then cd to its parent directory

    Remember current directory before the cd
        so that we can cd back there with cd('-')
    """
    if path_to == '-':
        if not cd.previous:
            raise PathError('No previous directory to return to')
        return cd(cd.previous)
    if not hasattr(path_to, 'cd'):
        path_to = makepath(path_to)
    try:
        previous = os.getcwd()
    except OSError as e:
        if 'No such file or directory' in str(e):
            return False
        raise
    if path_to.isdir():
        os.chdir(path_to)
    elif path_to.isfile():
        os.chdir(path_to.parent)
    elif not os.path.exists(path_to):
        return False
    else:
        raise PathError('Cannot cd to %s' % path_to)
    cd.previous = previous
    return True


try:
    cd.previous = makepath(os.getcwd())
except (OSError, AttributeError):
    cd.previous = None


def as_path(string_or_path):
    """Return the argument as a DirectPath

    If it is already one, return it unchanged
    If not, return the makepath()
    """
    if isinstance(string_or_path, DirectPath):
        return string_or_path
    return makepath(string_or_path)


def string_to_paths(string):
    for c in ':, ;':
        if c in string:
            return strings_to_paths(string.split(c))
    return [makepath(string)]


def strings_to_paths(strings):
    return [makepath(s) for s in strings]


def paths(strings):
    return [p for p in strings_to_paths(strings) if p.exists()]


def split_directories(strings):
    paths = strings_to_paths(strings)
    return [_
            for _ in paths
            if _.isdir()], [_ for _ in paths if not _.isdir()]


def split_files(strings):
    paths = strings_to_paths(strings)
    return ([_ for _ in paths if _.isfile()],
            [_ for _ in paths if not _.isfile()])


def split_directories_files(strings):
    paths = strings_to_paths(strings)
    return ([_ for _ in paths if _.isdir()],
            [_ for _ in paths if _.isfile()],
            [_ for _ in paths if not (_.isfile() or _.isdir())])


def files(strings):
    return split_files(strings)[0]


def directories(strings):
    return split_directories(strings)[0]


def home(sub_path = None):
    return makepath('~') / sub_path


def pwd():
    return makepath(os.getcwd())

here = pwd  # pylint: disable=invalid-name


def first_dir(path_string):
    """Get the first directory in that path

    >>> first_dir('usr/local/bin') == 'usr'
    True
    """
    parts = path_string.split(os.path.sep)
    return parts[0]


def first_dirs(path_strings):
    """Get the roots of those paths

    >>> first_dirs(['usr/bin', 'bin']) == ['usr', 'bin']
    True
    """
    return [first_dir(_) for _ in path_strings]


def unique_first_dirs(path_strings):
    """Get the unique roots of those paths

    >>> unique_first_dirs(['usr/local/bin', 'bin']) == set(['usr', 'bin'])
    True
    """
    return set(first_dirs(path_strings))


def _names_in_directory(path_to_directory):
    """Get all items in the given directory

    Swallow errors to give an empty list
    """
    try:
        return os.listdir(path_to_directory)
    except OSError:
        return []


def make_needed(pattern, path_to_directory, wanted):
    """Make a method to check if an item matches the pattern, and is wanted

    If wanted is None just check the pattern
    """
    if wanted:
        def needed(name):
            return fnmatch(name, pattern) and wanted(
                os.path.join(path_to_directory, name))
        return needed
    else:
        return lambda name: fnmatch(name, pattern)


def list_items(path_to_directory, pattern, wanted):
    """All items in the given path which match the given glob and are wanted"""
    if not path_to_directory:
        return set()
    needed = make_needed(pattern, path_to_directory, wanted)
    return [os.path.join(path_to_directory, name)
            for name in _names_in_directory(path_to_directory)
            if needed(name)]


def list_sub_directories(path_to_directory, pattern):
    """All sub-directories of the given directory matching the given glob"""
    return list_items(path_to_directory, pattern, os.path.isdir)


def set_items(path_to_directory, pattern, wanted):
    return set(list_items(path_to_directory, pattern, wanted))


def set_files(path_to_directory, pattern):
    """A list of all files in the given directory matching the given glob"""
    return set_items(path_to_directory, pattern, os.path.isfile)


def contains_glob(path_to_directory, pattern, wanted=None):
    """Whether the given path contains an item matching the given glob"""
    if not path_to_directory:
        return False
    needed = make_needed(pattern, path_to_directory, wanted)
    for name in _names_in_directory(path_to_directory):
        if needed(name):
            return True
    return False


def contains_directory(path_to_directory, pattern):
    """Whether the given path contains a directory matching the given glob"""
    return contains_glob(path_to_directory, pattern, os.path.isdir)


def contains_file(path_to_directory, pattern):
    """Whether the given directory contains a file matching the given glob"""
    return contains_glob(path_to_directory, pattern, os.path.isfile)


def environ_paths(key):
    return [makepath(_) for _ in os.environ[key].split(':')]


def environ_path(key):
    return makepath(os.environ[key])


def default_environ_path(key, default):
    return makepath(os.environ.get(key, default))


def tab_complete(string):
    """Finish file names "left short" by tab-completion

    For example, if an argument is "fred."
        and no file called "fred." exists
        but "fred.py" does exist
        then return fred.py
    """
    if is_option(string):
        return string
    if not missing_extension(string):
        return string
    if os.path.isfile(string):
        return string
    extended_files = [f for f in extend(string) if os.path.isfile(f)]
    try:
        return extended_files[0]
    except IndexError:
        return string


def pyc_to_py(path_to_file):
    """Change some file extensions to those which are more likely to be text

    >>> pyc_to_py('vim.pyc') == 'vim.py'
    True
    """
    stem, ext = os.path.splitext(path_to_file)
    if ext == '.pyc':
        return '%s.py' % stem
    return path_to_file



import os
import ast
import imp
from collections import defaultdict


from pysyte import dictionaries


class ImportVisitor(ast.NodeVisitor):
    def __init__(self):
        dd = defaultdict
        self.froms = dd(list)
        self.imports = dd(list)
        self.used = dd(list)
        super(ImportVisitor, self).__init__()

    def find_value_id(self, node, attr=None):
        if not node:
            return None
        value = getattr(node, 'value', None)
        if not value:
            if attr:
                value = getattr(getattr(node, attr, None), 'value', None)
                if not value:
                    name = getattr(getattr(node, attr, None), 'id', None)
                    if name:
                        return name
        if not value:
            return None
        result = getattr(value, 'id', None)
        if result:
            return result
        return self.find_value_id(value, attr)

    def find_name(self, node, *args):
        name = getattr(node, 'id', None)
        if name is not None:
            return name
        for attribute in args:
            attr = getattr(node, attribute, None)
            if attr:
                name = self.find_name(attr, *args)
                if name:
                    return name
        return None

    def check_usage(self, name, line):
        if not name:
            return
        found = name in self.imports
        if found:
            self.used[name].append(line)

    def unused(self):
        return {k: v for k, v in self.imports.items() if k not in self.used}

    def multiples(self):
        return {k: v for k, v in self.imports.items() if len(v) > 1}

    def collect_names(self, node):
        names = [(_.name, getattr(_, 'asname', None)) for _ in node.names]
        for name, alias in names:
            self.imports[alias if alias else name].append(node.lineno)
        return names

    def visit_ImportFrom(self, node):
        if node.module != '__future__':
            names = self.collect_names(node)
            self.froms[node.module].extend(names)
        self.generic_visit(node)

    def visit_Import(self, node):
        self.collect_names(node)
        self.generic_visit(node)

    def visit_FunctionDef(self, node):
        for decorator in node.decorator_list:
            name = self.find_name(decorator, 'func', 'value')
            self.check_usage(name, decorator.lineno)
        self.generic_visit(node)

    def visit_Subscript(self, node):
        name = self.find_value_id(node)
        name2 = self.find_name(node, 'value')
        assert name == name2
        self.check_usage(name, node.lineno)
        subscript = self.find_value_id(node.slice)
        name2 = self.find_name(node.slice, 'value')
        assert subscript == name2
        self.check_usage(subscript, node.lineno)
        self.generic_visit(node)

    def visit_Attribute(self, node):
        try:
            self.check_usage(node.value.id, node.lineno)
        except AttributeError:
            pass
        self.generic_visit(node)

    def visit_ClassDef(self, node):
        for base in node.bases:
            name = self.find_name(base, 'value')
            self.check_usage(name, node.lineno)
        self.generic_visit(node)

    def visit_Call(self, node):
        func = node.func
        try:
            name = func.id
        except AttributeError:
            name = self.find_value_id(func, 'func')
            name2 = self.find_name(node, 'value', 'func')
            assert name == name2
        self.check_usage(name, node.lineno)
        self.generic_visit(node)

    def visit_Name(self, node):
        name = self.find_name(node)
        self.check_usage(name, node.lineno)
        self.generic_visit(node)


def find_imports(tree):
    visitor = ImportVisitor()
    visitor.visit(tree)
    return visitor


def parse_python(path):
    with open(path) as stream:
        return ast.parse(stream.read(), path)


def extract_imports(script):
    """Extract all imports from a python script"""
    if not os.path.isfile(script):
        raise ValueError('Not a file: %s' % script)
    parse_tree = parse_python(script)
    result = find_imports(parse_tree)
    result.path = script
    return result


def load_module(package, module_name):

    def default_value(key):

        def imp_load(a, b, c):
            return imp.load_module(full_name, a, b, c)

        full_name = '%s.%s' % (package.__name__, module_name)
        return imp_load(*imp.find_module(key, package.__path__))

    # pylint: disable=protected-access
    try:
        _loaded = package._loaded
    except AttributeError:
        _loaded = package._loaded = dictionaries.LazyDefaultDict(default_value)
    return _loaded[module_name]



"""Provide methods to put colours in texts


Inspired by http://stackoverflow.com/questions/384076/how-can-i-make-the-python-logging-output-to-be-colored  # noqa
Conversion from RGB to xterm 256 colours based on values in http://www.frexx.de/xterm-256-notes/data/xterm-colortest  # noqa
Conversion from rgb.txt names to RGB inspired by https://github.com/lilydjwg/winterpy/blob/master/pyexe/gui2term.py  # noqa
"""


from __future__ import absolute_import
import os

from pysyte.colours import ansi_escapes
from pysyte.colours import colour_numbers


class ColouredTail(object):
    """The tail of a string that is being coloured

    Also holds a head, all of string before the tail
        When treated as a string, gives head + tail
    """
    def __init__(self, head, string):
        self._tail = string
        self._head = head

    def __repr__(self):
        return '<%s %r%r>' % (self.__class__.__name__, self.head, self.tail)

    def __str__(self):
        return ''.join((self.head, self.tail))

    def __eq__(self, other):
        return str(self) == str(other)

    @property
    def head(self):
        if self._head:
            return str(self._head)
        return ''

    @property
    def tail(self):
        if self._tail:
            return str(self._tail)
        return ''

    def colour_text(self, *args):
        new = colour_text(*args, head='')
        return ColouredTail(str(self), new)

    colour = text = colour_text


def colour_text(text, name=None, head=None):
    string = str(text)
    colour_ = colour_numbers.name_to_int(name) if name else False
    return ColouredTail(
        head or '',
        ansi_escapes.foreground_string(string, colour_) if colour_ else string
    )

colour = text = colour_text


def colour_initial(string, name):
    return colour(string[0], name).colour(string[1:])


def colour_initials(strings, name):
    return colour(''.join([str(colour_initial(s, name)) for s in strings]))


def prompt_text(string, name):
    number = colour_numbers.name_to_int(name)
    return ansi_escapes.prompt_string(string, number)



"""This module provides convenience methods to handle test files

Test files are expected to exist as triples:
    fred.py, fred.test, fred.tests
    with fred.test* being optional

We also allow for python scripts which have no extension
"""


import os
import re
import sys
import argparse


try:
    from pysyte.paths import path
except ImportError:
    from dotsite.paths import path


class UserMessage(Exception):
    """Tell the user something went wrong"""
    pass

class MissingFiles(UserMessage):
    pass

def first_thing_that(things, predicate):
    for thing in things:
        if predicate(thing):
            return thing
    return None


def _get_path_to_test_directory(strings):
    """Get a path to the root directory that we'll be testing in

    We will test in the first directory in strings
        or the directory of the first file in strings
    """
    paths = [path(s) for s in strings]
    result = first_thing_that(paths, lambda p: p.isdir())
    if not result:
        try:
            result = first_thing_that(paths, lambda p: p.parent and p.isfile()).parent
        except AttributeError:
            raise MissingFiles('No doctests found in %r' % strings)
    if not result.files('*.test*') and not result.files('*.py'):
        raise MissingFiles('No doctests found in %r' % strings)
    return result


def _existing_test_files(path_to_stem):
    """A list of all existing extensions of the given path

    Find all such extensions by replacing any extension in a real file
        or adding "/*" if it is a directory
        or otherwise looking the current directory

    Not expecting any test scripts in the root directory
    >>> _existing_test_files(path('/test_files.txt')) == []
    True

    But we should expect them for this script
    >>> __file__ in _existing_test_files(path(__file__))
    True
    """
    if path_to_stem.isfile():
        dirr = path_to_stem.parent
        glob = path_to_stem.namebase + '.*'
    elif path_to_stem.isdir():
        dirr = path_to_stem
        glob = '*'
    else:
        if path_to_stem.parent:
            dirr = path_to_stem.parent
        glob = '%s.*' % path_to_stem.namebase
    if not dirr.isdir():
        return []
    return [f for f in dirr.files(glob)]


def _existing_test_extensions(path_stem):
    """A list of all text extensions of the given path stem"""
    return [f for f in _existing_test_files(path_stem) if f.ext in possible_test_extensions()]


def _get_path_stems(strings, recursive):
    """A list of any paths in strings

    If there is a "/" in the strings then it is used in the list
    Otherwise it is appended to the current working directory
    """
    strings = strings or []
    here = path('.')
    paths = ['/' in s and path(s) or here / s for s in strings]
    if not recursive:
        return paths
    return add_sub_dirs(paths)


def add_sub_dirs(paths):
    """Add all sub-directories for the directories of the paths"""
    dirs = {p.directory() for p in paths}
    result = dirs.copy()
    for path_to_dir in dirs:
        for sub_dir in path_to_dir.walkdirs():
            if '.git' in sub_dir.splitall():
                continue
            result.add(sub_dir)
    return result


def python_source_extensions():
    """The extensions of files that often contain python source"""
    return ['.py', '']


def python_doctest_extensions():
    """The extensions of files that often contain python doctests"""
    return ['.tests', '.test']


def possible_test_extensions():
    """All possible extensions for python files and test files"""
    return python_doctest_extensions() + python_source_extensions()


def positive_test_extensions():
    """Positive extensions for python files and test files"""
    return [ext for ext in possible_test_extensions() if ext]


def is_python_source_extension(string):
    """Whether the string is an extension for a file with python source"""
    return string in python_source_extensions()


def is_python_doctest_extension(string):
    """Whether the string is an extension for a file with python doctests"""
    return string in python_doctest_extensions()


def _is_possible_test_extension(string):
    """Whether the string is a possible test extension"""
    return string in possible_test_extensions()


def _is_positive_test_extension(string):
    """Whether the string is a positive test extension"""
    return string in positive_test_extensions()


def _has_checked_extension(string, checker):
    """Whether the string has an extension validated by checker"""
    _, ext = os.path.splitext(string)
    return checker(ext)


def has_python_source_extension(string):
    """Whether the string has an extension for a file with python source"""
    return _has_checked_extension(string, is_python_source_extension)


def has_python_doctest_extension(string):
    """Whether the string has an extension for a file with python doctests"""
    return _has_checked_extension(string, is_python_doctest_extension)


def _has_possible_test_extension(string):
    """Whether the string has a possible test extension"""
    return _has_checked_extension(string, _is_possible_test_extension)


def _has_positive_test_extension(string):
    """Whether the string has a possible test extension"""
    return _has_checked_extension(string, _is_positive_test_extension)


def _positive_test_globs():
    """A glob for each of the _positive test extensions"""
    return [str('*%s' % ext) for ext in positive_test_extensions()]


def has_doctests(string):
    """Whether the string contains doctests

    This method actually checks for presence of the classic "    >>> " prefix

    >>> has_doctests('    >>> print True')
    True
    >>> has_doctests('not space    >>> print True')
    False
    """
    doctest_regexp = re.compile(r'^\s*>>>\s', re.MULTILINE)
    return bool(doctest_regexp.search(string))


def _all_possible_test_files_in(path_to_root, recursive):
    """A list of all possible test files in the given root

    If recursive is True then include sub-directories
    """
    find_files = recursive and path_to_root.walkfiles or path_to_root.listfiles
    result = []
    ignores = ['.git/', '.svn/', '.idea/',
               'coverage/', 'build/', '*.egg-info/', 'dist/', '__pycache__/', '.tox/',
               '*.sw[op]']
    for glob in _positive_test_globs():
        files = find_files(glob, ignores=ignores)
        result.extend(files)
    ignores.extend(_positive_test_globs())
    for path_to_file in find_files(ignores=ignores):
        if path_to_file.hidden:
            continue
        try:
            if not path_to_file.ext and _first_line_is_python_shebang(path_to_file.stripped_lines()):
                result.append(path_to_file)
        except UnicodeDecodeError as e:
            raise ValueError('%s - %s' % (path_to_file, repr(e)))
    return result


def _get_scripts_here(recursive):
    """Find all test scripts in the current working directory"""
    here = path('.')
    all_possibles = _all_possible_test_files_in(here, recursive)
    test_extensions = [_ for _ in all_possibles if has_python_doctest_extension(_)]
    test_texts = [_ for _ in all_possibles if has_python_source_extension(_) and has_doctests(_.text())]
    return test_texts + test_extensions


def _expand_stems(path_stems):
    """Extend the list of path stems with all test-extensions"""
    result = []
    for path_stem in path_stems:
        if path_stem.isfile():
            result.append(path_stem)
        else:
            result.extend(_existing_test_extensions(path_stem))
    if not result:
        path_stem = path_stems.pop()
        display_stem = path_stem.short_relative_path_from_here()
        raise MissingFiles('%s.test*, %s.py not found' % (display_stem, display_stem))
    return result


def _get_files_from_stems(strings, recursive):
    """Get a list of test scripts from the given strings

    Only known extensions are included
    If recursive is true then include sub_directories
    """
    path_stems = _get_path_stems(strings, recursive)
    if path_stems:
        return _expand_stems(path_stems)
    return _get_scripts_here(recursive)


def _first_line_is_python_shebang(lines):
    """Whether the first line of that script looks like

    #! ...python...
    """
    if not lines:
        return False
    first_line = lines[0]
    return bool(re.match('#!.*python.*', first_line))


def _re_order_scripts(paths_to_scripts):
    """Re order the list of scripts so that their extensions are in order

    The order to be used is same as the extensions of possible_test_extensions()
        which goes from more testy to more doccy
    """
    result = []
    for extension in possible_test_extensions():
        extension_paths = [p for p in paths_to_scripts if p.ext == extension]
        result.extend(extension_paths)
    return result


def paths_to_doctests(strings, recursive):
    """Get paths to all doctest files for the given strings

    strings should be a list of stems - possibly empty
        A "stem" is the start of a path, e.g. "test_files."

    If recursive is True then include sub_directories
    """
    paths_to_files = _get_files_from_stems(strings, recursive)
    paths_to_positive_scripts = [p for p in paths_to_files if p.ext in positive_test_extensions()]
    for path_to_script in paths_to_files:
        if path_to_script in paths_to_positive_scripts:
            continue
        if path_to_script.ext:
            continue
        if not _first_line_is_python_shebang(path_to_script.stripped_lines()):
            continue
        paths_to_positive_scripts.append(path_to_script)
    return _re_order_scripts(paths_to_positive_scripts)


def handle_command_line():
    """Find options and arguments on the command line"""
    parser = argparse.ArgumentParser(description=__doc__.splitlines()[0])
    parser.add_argument(
        'stems', type=str, nargs='*', help='filename stems')
    parser.add_argument(
        '-r', '--recursive', action='store_true', help='Look in sub-directories')
    args = parser.parse_args()
    stems = args.stems if args.stems else [os.path.dirname(__file__)]
    return stems, args.recursive


def main():
    """Run the progam"""
    stems, recursive = handle_command_line()
    try:
        scripts = paths_to_doctests(stems, recursive)
    except UserMessage as e:
        print(e, file=sys.stderr)
        return 1
    print('\n'.join(scripts))
    return 0


if __name__ == '__main__':
    sys.exit(main())



"""Script to handle running of doctests"""


import re
import os
import io
import sys
import importlib
import doctest
import getpass
import socket
import subprocess
import datetime
from pprint import pprint


from pysyte.paths import path

from pysyte.testing import files_for_test
from pysyte.testing import try_plugins
from pysyte.testing.see import see, see_methods, see_attributes, spread

__version__ = '0.7.3.1'

class DoctestInterrupt(KeyboardInterrupt):
    """This exception is for better naming of the only thing to stop doctest"""


def no_print(method, *args, **kwargs):
    """Discard all writes to stdout"""
    old_stdout = sys.stdout
    try:
        sys.stdout = io.StringIO()
        return method(*args, **kwargs)
    finally:
        sys.stdout = old_stdout


def run_command(command):
    """Run a command in the local shell (usually bash)"""
    status, output = subprocess.getstatusoutput(command)
    if status:
        print('FAIL: %s:%s' % (status, output))
        return False
    print(output)
    return True


def show(thing):
    """Pretty print the given thing

    Unless the thing is a module
        when we show the help for that module instead
    """
    if isinstance(thing, type(re)):
        return pprint(help(thing))
    return pprint(thing)


def get_host():
    return os.environ.get('HOST', os.environ.get('HOSTNAME',
        socket.getfqdn().split('.')[0]))


def get_user():
    return getpass.getuser()


class TestBeingRun(object):
    """Encapsulation of the current test"""

    def __init__(self, that):
        self.here = path('.')
        self.home = path('~')
        self.username = get_user()
        self.host = get_host()
        self.user = '@'.join((self.username, self.host))
        self.path = self.here.relpathto(that)
        base, _ext = self.path.splitext()
        [self.add_ext(base, _) for _ in {_ext[1:], 'py', 'test', 'tests', 'fail'}]

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, str(
            self.path.short_relative_path_from_here()))

    def add_ext(self, base, ext):
        """Add a path for base.ext as an attribute to self"""
        name = 'path_to_%s' % ext
        if hasattr(self, name):
            return
        path_to_ext = path('%s.%s' % (base, ext))
        if path_to_ext.isfile() or ext == 'fail':
            setattr(self, name, path_to_ext)


def parse_args():
    """Handle args on the command line

    Destroy the option parser when finished
        (do not interfere with tests which use it)
    """
    # Hello future me , please convert to argparse
    import argparse
    parser = argparse.ArgumentParser(description=__doc__.splitlines()[0])
    pa = parser.add_argument
    pa('stems', metavar='stems', type=str, nargs='*',
       help='stems to be tested (e.g "try.py" or "try.*" or "try." or "try/"')
    pa('-s', '--show', help='show files being tested', action='store_true')
    pa('-v', '--verbose', help='Show more text', action='store_true')
    pa('-V', '--version', help='Show version', action='store_true')
    pa('-r', '--recursive', action='store_true',
       help='recurse into any sub-directories found',)
    pa('-a', '--all', action='store_true',
       help="run all tests (don't stop on first FAIL)",)
    pa('-d', '--directory_all', action='store_true',
       help='run all test scripts in a directory'
       '(do not stop on first FAILing script)',)
    pa('-q', '--quiet_on_success', action='store_true',
       help='no output if all tests pass',)
    pa('-p', '--persist', action='store_true',
       help='do not stop on DoctestInterrupts',)
    args = parser.parse_args()
    # Try not to interfere with any ArgumentParser used in script under test
    if hasattr(parser, 'destroy'):
        parser.destroy()
    del parser
    if args.recursive:
        stems = [_ for _ in args.stems if os.path.isfile(_)]
        if not stems:
            return args
        raise ValueError(
            'Do not use --recursive with stems ("%s")' % (
            '"%s", '.join(stems)))
    return args


class SysPathHandler(object):

    def __init__(self):
        self.paths = []

    def add(self, item):
        """Add the item to sys.path"""
        directory = path(item).directory()
        if directory not in self.paths:
            self.paths.insert(0, directory)
            if directory not in sys.path:
                sys.path.insert(0, str(directory))

    def remove(self, item):
        """Remove the item from sys.path"""
        directory = path(item).directory()
        if directory in self.paths:
            self.paths.remove(directory)
            if directory in sys.path:
                sys.path.remove(directory)


def make_module(path_to_python):
    """The module object for the given python source code

    https://stackoverflow.com/a/67692/500942
    """
    name = path_to_python.namebase
    try:
        return sys.modules[name]
    except KeyError:
        spec = importlib.util.spec_from_file_location(name, path_to_python)
        return importlib.util.module_from_spec(spec)


def make_python2_module(path_to_python):
    try:
        return sys.modules[name]
    except KeyError:
        try:
            fp, pathname, description = importlib.find_module(
                name, [path_to_python.parent])
        except ImportError:
            try:
                fp, pathname, description = (
                    open(name), path_to_python, ('', 'r', importlib.PY_SOURCE))
            except IOError:
                raise ImportError('Could not find a module for %r' %
                                  str(path_to_python))
        # If any of the following calls raises an exception,
        # there's a problem we can't handle -- let the caller handle it.
        try:
            return importlib.load_module(name, fp, pathname, description)
        finally:
            if fp:
                fp.close()


def test_source(source_script):
    """Run tests in the given source"""
    message = 'py %s;' % source_script
    module = make_module(source_script)
    failures, tests_run = doctest.testmod(
        module,
        optionflags=get_doctest_options(),
        verbose=args.verbose,
    )
    del module
    return failures, tests_run, message


def test_file(test_script):
    """Run tests in a doctest script"""
    message = 'try %s;' % test_script
    failures, tests_run = doctest.testfile(
        test_script,
        optionflags=get_doctest_options(),
        module_relative=False,
        globs={
            'test': TestBeingRun(test_script),
            'sys': sys,
            'see': see,
            'pp': pprint,
            'spread': spread,
            'see_methods': see_methods,
            'see_attributes': see_attributes,
            'makepath': path,
            'show': show,
            'bash': run_command,
            'DoctestInterrupt': DoctestInterrupt,
            'no_print': no_print,
        },
        verbose=args.verbose,
    )
    return failures, tests_run, message


def run_doctest(test_script):
    """Call the relevant doctest method for the given script"""
    if test_script.ext in ['', '.py']:
        try:
            return test_source(test_script)
        except ImportError:
            if test_script.ext:
                raise
            return 0, 0, ''
    return test_file(test_script)


def show_running_doctest(test_script):
    """Run a doctest for that script, showing progress on stdout/stderr"""
    old_argv = sys.argv[:]
    try:
        sys.argv = [test_script]
        try:
            if args.verbose or args.show:
                print('Test', test_script)
            return run_doctest(test_script)
        except DoctestInterrupt as e:
            if args.directory_all or args.persist:
                show_interruption(test_script, e)
                return 0, 0, ''
            raise
        except SyntaxError as e:
            if args.directory_all or args.persist:
                show_exception(test_script, e)
                return 0, 0, ''
            raise
    finally:
        sys.argv[:] = old_argv


def get_doctest_options():
    """Convert the command line args to Doctest flags"""
    result = doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE
    if not args.all:
        result |= doctest.REPORT_ONLY_FIRST_FAILURE
    return result


def show_time_taken(start, messages, message, tests_run):
    """Add a message about the time taken dince start to the messages"""
    end = datetime.datetime.now()
    time_taken = end - start
    if time_taken.seconds:
        if time_taken.seconds > 1:
            time_msg = 'in %d seconds' % time_taken.seconds
        else:
            time_msg = 'quickly'
    else:
        time_msg = 'very quickly'
    messages.append('%s %s tests passed %s' % (message, tests_run, time_msg))
    return end


def show_exception(test_script, exception):
    """Show the reason for an interuption on stderr"""
    print('Bye from ', test_script, file=sys.stderr)
    print('Because:', interruption, file=sys.stderr)


def show_interruption(*args):
    """Shout ^C back to stderr"""
    c_line = ('^c ^C ' * 4) + '\n'
    print(c_line * 3, file=sys.stderr)
    show_exception(*args)


args = None

def test():
    """Run all tests"""
    sys_paths = SysPathHandler()
    global args
    args = parse_args()
    if args.version:
        print(__version__)
        return 0
    messages = ['']
    pwd = os.getcwd()
    end = start_all = datetime.datetime.now()
    run_all = 0
    failures_all = 0
    sys_paths.add('.')
    try:
        for test_script in files_for_test.paths_to_doctests(
                args.stems, args.recursive):
            failures, tests_run, message = 0, 0, ''
            os.chdir(pwd)
            start = datetime.datetime.now()
            try:
                sys_paths.add(test_script)
                if try_plugins.pre_test(test_script):
                    failures, tests_run, message = show_running_doctest(
                        test_script)
                    try_plugins.post_test(test_script, failures, tests_run)
            finally:
                sys_paths.remove(test_script)
            if tests_run:
                run_all += tests_run
                end = show_time_taken(start, messages, message, tests_run)
            else:
                show_time_taken(start, messages, message, tests_run)
            if failures:
                failures_all += failures
                if not args.directory_all:
                    return failures
    finally:
        time_taken = end - start_all
        messages.append('%s tests passed, %d failed, in %s seconds' %
                        (run_all, failures_all, time_taken.seconds))
        messages.append('')
    if failures_all or not args.quiet_on_success:
        print('\n'.join(messages))
    return failures_all



"""This module holds plugins for try.py"""



def pre_test(_path_to_test_file):
    """This method is called before each test has been run

    If this method returns a false value then that file is not tested
    """
    return True


def post_test(_path_to_test_file, _failures, _tests_run):
    """This method is called after each test has been run"""
    return True



"""This module provides methods to handle some features of English words"""


import re


import inflect

_inlection = inflect.engine()


def pluralize(string):
    """Returns the plural of string.

    >>> pluralize('Cow')  == 'Cows'
    True
    """
    return _inlection.plural_noun(string)


def singularize(string):
    """Returns the singular of string.

    >>> singularize('Cows') == 'Cow'
    True
    """
    return _inlection.singular_noun(string)


def number_name(item):
    """The English name for that number

    >>> assert number_name(5) == 'five'
    """
    return _inlection.number_to_words(item)



"""Facilities for tracebacks"""

import re


def line_regexp():
    """Regular expression to match a traceback file line"""
    return re.compile(r'''\s*
        File\s
        (
            (
                ["']
                (?P<path_to_python>[^"']+)
                ["']
            ) | (
                (?P<spaceless_path_to_python>[^ ]+)
            )
        )
        ,\sline\s
        (?P<line_number>[0-9]+)
        ,.in..*
    ''', re.VERBOSE)


def parse_line(string):
    """Parse a single string as traceback line"""
    match = line_regexp().match(string)
    if match:
        matches = match.groupdict()
        line_number = matches['line_number']
        path_to_python = matches['path_to_python']
        spaceless_path_to_python = matches['spaceless_path_to_python']
        if path_to_python:
            return path_to_python, line_number
        elif spaceless_path_to_python:
            return spaceless_path_to_python, line_number



"""Proxies for known debuggers"""

import inspect
from bdb import BdbQuit as DebugExit  # pylint: disable=unused-import


try:
    from pprintpp import pprint as pp
except ImportError:
    from pprint import pprint as pp


def ppd(_):
    return pp(dir(_))


def ppv(_):
    return pp(vars(_))


def source_of(frame):
    return frame.f_code.co_filename, frame.f_lineno


class Debugger(object):
    """Proxy to a debugger"""
    def _continue(self, frame):
        raise NotImplementedError('To do: running in de other buggers')

    def save_breaks(self, breakpoints):
        raise NotImplementedError('To do: breakpoints in de other buggers')

    def load_breaks(self):
        raise NotImplementedError('To do: breakpoints in de other buggers')

    def _break_on(self, frame):
        filename, lineno = source_of(frame)
        self.add_break(filename, lineno)

    def break_here(self, frame):
        self._break_on(frame)
        self._continue(frame)

    def add_break(self, filename, lineno):
        self.add_breaks([(filename, lineno)])

    def _make_break(self, filename, lineno):  # pylint: disable=no-self-use
        return (filename, lineno)

    def add_breaks(self, filenames_lines):
        self.save_breaks(
            self.load_breaks() + [
                self._make_break(f, l) for f, l in filenames_lines
            ])


class PythonDebugger(Debugger):
    def _continue(self, frame):
        raise NotImplementedError('To do: running in de other buggers')

    def save_breaks(self, breakpoints):
        raise NotImplementedError('To do: breakpoints in de other buggers')

    def load_breaks(self):
        raise NotImplementedError('To do: breakpoints in de other buggers')


class PudbDebugger(Debugger):
    """Proxy to the pudb debugger"""
    @property
    def debugger(self):
        try:
            return pudb.CURRENT_DEBUGGER[0]
        except IndexError:
            return None

    def _make_break(self, filename, lineno):
        return (filename, lineno, False, None, None)

    def _continue(self, frame):
        if self.debugger:
            pudb.set_interrupt_handler()
            self.debugger.set_trace(frame)
        else:
            pudb.set_trace()

    def break_here(self, frame):
        # pylint: disable=protected-access
        if self.debugger:
            self.debugger.break_here(frame)
        else:
            dbg = pudb._get_debugger()
            import threading
            if isinstance(threading.current_thread(), threading._MainThread):
                pudb.set_interrupt_handler()
            dbg.set_trace(frame)

    def save_breaks(self, breakpoints):
        class PudbBreakpoint(object):
            """Cadged from .../pudb/settings.py

            https://github.com/inducer/pudb/blob/master/pudb/settings.py#L494
            """
            def __init__(self, filename, lineno, condition):
                self.file = filename
                self.line = lineno
                self.cond = condition

        save_breakpoints = [PudbBreakpoint(*b[:3]) for b in breakpoints]
        pudb.settings.save_breakpoints(save_breakpoints)

    def load_breaks(self):
        return pudb.settings.load_breakpoints()

    def add_breaks(self, filenames_lines):
        super(PudbDebugger, self).add_breaks(filenames_lines)
        if self.debugger:
            self.debugger.reload_breakpoints(0, 0, 0)  # args not used


class IPythonDebugger(object):
    """Proxy to ipdb's debugger"""
    def _continue(self, frame):
        raise NotImplementedError('To do: running in de other buggers')

    def save_breaks(self, breakpoints):
        raise NotImplementedError('To do: breakpoints in de other buggers')

    def load_breaks(self):
        raise NotImplementedError('To do: breakpoints in de other buggers')


try:
    import pudb
    db = PudbDebugger()
except ImportError:
    try:
        # pylint: disable=redefined-variable-type
        db = IPythonDebugger()
    except ImportError:
        db = PythonDebugger()


def debug_here(really=True):
    if not really or not db:
        return
    frame = inspect.currentframe().f_back
    db.break_here(frame)


def stack_sources():
    """A list of sources for frames above this"""
    # lazy imports
    import linecache
    result = []
    for frame_info in reversed(inspect.stack()):
        _frame, filename, line_number, _function, _context, _index = frame_info
        linecache.lazycache(filename, {})
        _line = linecache.getline(filename, line_number).rstrip()

    # Each record contains a frame object, filename, line number, function
    # name, a list of lines of context, and index within the context
    _sources = [(path, line) for _, path, line, _, _, _ in inspect.stack()]
    return result



"""Methods to provide names for colours

And numbers for those names"""


def bw():
    """The simplest of all - no colour and full colour"""
    return 'black white'.split()


def rgb():
    """The basic cone sensitivities of your eyes"""
    return 'red green blue'.split()


primaries = rgb


def cym():
    """The complements of the primaries"""
    return 'cyan yellow magenta'.split()


secondaries = complementaries = cym


def basic():
    """The basic colour names: Black, the primaries, their complements and white"""
    return bw() + rgb() + cym()


def _cga_colour_names():
    """The basic CGA colours are in a slightly different order"""
    return 'red green yellow blue magenta cyan'.split()


def dark_cga():
    return ['black'] + _cga_colour_names() + ['silver']


def light_cga():
    return ['gray'] + [str('light %s' % name) for name in _cga_colour_names()]  + ['white']


def cga():
    return dark_cga() + light_cga()



"""Generic methods for splitting strings"""


import re

from pysyte import literals


def _default_separator():
    """The default separator used by all methods is a comma

    Monkey-patching this method can change that default
    """
    return literals.Punctuation.comma


def join(items, separator=None):
    """Join the items into a string using the separator

    Converts items to strings if needed

    >>> join([1, 2, 3])
    '1,2,3'
    """
    if not items:
        return ''
    if separator is None:
        separator = _default_separator()
    return separator.join([str(item) for item in items])


def seamless_join(items):
    """A seamless join does not show where the joins are

    >>> seamless_join(['fred', 'was', 'here'])
    'fredwashere'
    """
    return join(items, literals.Empty.string)


def split(string, separator_regexp=None, maxsplit=0):
    """Split a string to a list

    >>> split('fred, was, here')
    ['fred', ' was', ' here']
    """
    if not string:
        return []
    if separator_regexp is None:
        separator_regexp = _default_separator()
    if not separator_regexp:
        return string.split()
    return re.split(separator_regexp, string, maxsplit)


def split_and_strip(string, separator_regexp=None, maxsplit=0):
    """Split a string into items and trim any excess spaces from the items

    >>> split_and_strip('fred, was, here  ')
    ['fred', 'was', 'here']
    """
    if not string:
        return ['']
    if separator_regexp is None:
        separator_regexp = _default_separator()
    if not separator_regexp:
        return string.split()
    return [item.strip()
            for item in re.split(separator_regexp, string, maxsplit)]


def split_and_strip_without(string, exclude, separator_regexp=None):
    """Split a string into items, and trim any excess spaces

    Any items in exclude are not in the returned list

    >>> split_and_strip_without('fred, was, here  ', ['was'])
    ['fred', 'here']
    """
    result = split_and_strip(string, separator_regexp)
    if not exclude:
        return result
    return [x for x in result if x not in exclude]


def split_and_strip_whole(string, separator_regexp=None):
    """Split a string into items and trim any excess spaces from the items

    Exclude any empty items

    >>> split_and_strip_whole('fred, , was,here,')
    ['fred', 'was', 'here']
    """
    return split_and_strip_without(string, [''], separator_regexp)


def split_by_count(items, count, filler=None):
    """Split the items into tuples of count items each

    >>> split_by_count([0,1,2,3], 2)
    [(0, 1), (2, 3)]

    If there are a mutiple of count items then filler makes no difference
    >>> split_by_count([0,1,2,7,8,9], 3, 0) == split_by_count([0,1,2,7,8,9], 3)
    True

    If there are not a multiple of count items, then any extras are discarded
    >>> split_by_count([0,1,2,7,8,9,6], 3)
    [(0, 1, 2), (7, 8, 9)]

    Specifying a filler expands the "lost" group
    >>> split_by_count([0,1,2,7,8,9,6], 3, 0)
    [(0, 1, 2), (7, 8, 9), (6, 0, 0)]
    """
    if filler is not None:
        items = items[:]
        while len(items) % count:
            items.append(filler)
    iterator = iter(items)
    iterators = [iterator] * count
    return list(zip(*iterators))


def pairs(items, filler=None):
    """Split the items into pairs

    >>> pairs([0,1,2,7,8,9,10])
    [(0, 1), (2, 7), (8, 9)]

    If a filler is used any unpaired items are retained
    >>> pairs([0,1,2,7,8,9,10], 99)
    [(0, 1), (2, 7), (8, 9), (10, 99)]
    """
    return split_by_count(items, 2, filler)


def threes(items, filler=None):
    """Split the items into groups of 3

    >>> threes([0,1,2,6,7,8,9])
    [(0, 1, 2), (6, 7, 8)]

    If a filler is used any discarded items are retained
    >>> threes([0,1,2,6,7,8,9], 5)
    [(0, 1, 2), (6, 7, 8), (9, 5, 5)]
    """
    return split_by_count(items, 3, filler)


def despaced(string):
    """Split a string into spaceless items

    Split on spaces, trim excess space, exclude any empty strings

    >>> despaced('fred, , was,here today')
    ['fred,', ',', 'was,here', 'today']
    """
    return split_and_strip_without(string, [''], ' ')


def words(string):
    """Split a string into words

    Split on (English) punctuaution, trim space, exclude any empty strings

    >>> words('fred, , was,here today')
    ['fred', 'was', 'here', 'today']
    """
    return split_and_strip_without(string, [''], '[,;. ]')


def rejoin(string, separator_regexp=None, spaced=False):
    """Split a string and then rejoin it

    Spaces are interspersed between items only if spaced is True

    >>> rejoin('fred, was, here  ')
    'fred,was,here'
    """
    strings = split_and_strip(string)
    if separator_regexp is None:
        separator_regexp = _default_separator()
    joiner = spaced and '%s ' % separator_regexp or separator_regexp
    return joiner.join(strings)



import platform as python_platform

import pysyte
from pysyte import imports
from pysyte.bash import cmnds

_platform_name = python_platform.system().lower()
platform = imports.load_module(pysyte, _platform_name)


def _bash(command):
    if not command:
        return None
    status, output = cmnds.getstatusoutput(command)
    if status:
        raise ValueError(output)
    return output


def get_clipboard_data():
    return _bash(platform.bash_paste)



"""Provide a getch() method to get single chars from a terminal


Do not wait on carriage returns

Adapted from
    http://code.activestate.com/recipes/134892/
    Without the redundant classes of that code
        but keeping the implementatation

    And reducing down to just the tty version
        No need here for Windows/Carbon
"""

import re
import signal
import sys
import tty
from curses import ascii

import termios


class NoKeys(StopIteration):
    """Better name for StopIteration caused by running out of keys"""
    pass


def get_ord():
    """The integer ordinal of the next byte read from sys.stdin"""
    return ord(sys.stdin.read(1))


class TerminalContext(object):
    """Context wrapper to set up termios values"""
    # pylint: disable=too-few-public-methods
    def __init__(self):
        self.fd = None
        self.old_settings = None

    def __enter__(self):
        self.fd = sys.stdin.fileno()
        self.old_settings = termios.tcgetattr(self.fd)
        mode = termios.tcgetattr(self.fd)
        mode[tty.LFLAG] = mode[tty.LFLAG] & ~(termios.ECHO | termios.ICANON)
        termios.tcsetattr(self.fd, termios.TCSAFLUSH, mode)
        return self

    def __exit__(self, typ, value, traceback):
        if typ is not None:
            pass  # exception
        if self.old_settings:
            termios.tcsetattr(self.fd, termios.TCSADRAIN, self.old_settings)


class Timeout(Exception):
    """A timeout has occurred"""
    pass


def timeout_raiser(_signum, _frame):
    """Raise a timeout exception"""
    raise Timeout()


class TimerContext(object):
    """Context wrapper for a timeout"""
    # pylint: disable=too-few-public-methods
    def __init__(self, seconds):
        self.seconds = seconds
        self.old_handler = None
        self.timed_out = False

    def __enter__(self):
        self.old_handler = signal.signal(signal.SIGALRM, timeout_raiser)
        signal.setitimer(signal.ITIMER_REAL, self.seconds)
        return self

    def __exit__(self, typ_, value, traceback):
        signal.setitimer(signal.ITIMER_REAL, 0)
        if self.old_handler:
            signal.signal(signal.SIGALRM, self.old_handler)
        if isinstance(value, Timeout):
            self.timed_out = True
            return True


_key_cache = []


def cache_keys(keys):
    """Allow debugging via PyCharm"""
    d = known_keys()
    known_names = dict(zip(d.values(), d.keys()))
    for k in keys:
        i = (ord(k),) if len(k) == 1 else known_names[k]
        _key_cache.insert(0, i)


def _get_keycodes():
    """Read keypress giving a tuple of key codes

    A 'key code' is the ordinal value of characters read

    For example, pressing 'A' will give (65,)
    """
    try:
        return _key_cache.pop()
    except IndexError:
        pass
    result = []
    terminators = 'ABCDFHPQRS~'
    with TerminalContext():
        code = get_ord()
        result.append(code)
        if code == 27:
            with TimerContext(0.1) as timer:
                code = get_ord()
            if not timer.timed_out:
                result.append(code)
                result.append(get_ord())
                if 64 < result[-1] < 69:
                    pass
                elif result[1] == 91:
                    while True:
                        code = get_ord()
                        result.append(code)
                        if chr(code) in terminators:
                            break
    return tuple(result)


class ExtendedKey(Exception):
    """Raised for an unrecognised Extended key code"""
    def __init__(self, codes):
        super(ExtendedKey, self).__init__(
            'Too many key codes: %s' % repr(codes))
        self.codes = codes


def getch():
    """Get a char or and ExtendedKey from a keyboard"""
    codes = _get_keycodes()
    if len(codes) == 1:
        return chr(codes[0])
    raise ExtendedKey(codes)


def get_codes():
    """Get a tuple of stringified keycodes from the keyboard"""
    return tuple(str(_) for _ in _get_keycodes())


def get_key():
    """Get a key from the keyboard as a string

    A 'key' will be a single char, or the name of an extended key
    """
    character_name = chr
    codes = _get_keycodes()
    if len(codes) == 1:
        code = codes[0]
        if code >= 32:
            return character_name(code)
        return control_key_name(code)
    return get_extended_key_name(codes)


def get_menu(**kwargs):
    key = get_key()
    for name, string in kwargs.items():
        match = re.match('^%s$' % string, key)
        if match:
            return name
        if key in name or key in string:
            return name
    return key


def get_ascii():
    """Get ASCII key from stdin

    return None for others
    """
    try:
        return getch()
    except ExtendedKey:
        return None


def get_ASCII():
    """Get ASCII key from stdin

    raise error on other
    """
    try:
        return getch()
    except ExtendedKey:
        raise KeyboardInterrupt


def get_as_key():
    """Get key from stdin

    return ASCII keys as (single char) strings
    others as tuples
    """
    try:
        return getch()
    except ExtendedKey as e:
        return e.codes


def control_key_name(code):
    """Prefix the name of a control key with '^'"""
    return '^%s' % (chr(code - 1 + ord('A')))


def get_extended_key_name(codes):
    """Get a name for the extended key with those codes

    Names are defined herein
    """
    return known_keys()[codes]


def known_keys():
    return {
        (27, 79, 70): 'end',
        (27, 79, 72): 'home',
        (27, 79, 80): 'F1',
        (27, 79, 81): 'F2',
        (27, 79, 82): 'F3',
        (27, 79, 83): 'F4',
        (27, 91, 49, 53, 126): 'F5',
        (27, 91, 49, 55, 126): 'F6',
        (27, 91, 49, 56, 126): 'F7',
        (27, 91, 49, 57, 126): 'F8',
        (27, 91, 50): 'insert',
        (27, 91, 50, 48, 126): 'F9',
        (27, 91, 50, 49, 126): 'F10',
        (27, 91, 50, 51, 126): 'F11',
        (27, 91, 50, 52, 126): 'F12',
        (27, 91, 49, 59, 50, 80): 'F13',
        (27, 91, 49, 59, 50, 83): 'F16',
        (27, 91, 49, 53, 59, 50, 126): 'F17',
        (27, 91, 49, 55, 59, 50, 126): 'F18',
        (27, 91, 49, 56, 59, 50, 126): 'F19',
        (27, 91, 51, 126): 'Del',
        (27, 91, 51): 'delete',
        (27, 91, 53): 'page up',
        (27, 91, 53, 126): 'page up',
        (27, 91, 54): 'page down',
        (27, 91, 54, 126): 'page down',
        (27, 91, 65): 'up',
        (27, 91, 66): 'down',
        (27, 91, 67): 'right',
        (27, 91, 68): 'left',
    }


def _yielder(getter):
    """Keep yielding from that getter until KeyboardInteruptted"""
    def yield_till_stopped():
        """Yield from the getter until KeyboardInteruptted"""
        while True:
            try:
                yield getter()
            except KeyboardInterrupt:
                raise NoKeys

    return yield_till_stopped


yield_keycodes = _yielder(_get_keycodes)
yield_asciis = _yielder(get_ascii)
yield_ASCIIs = _yielder(get_ASCII)
yield_as_keys = _yielder(get_as_key)


def get_string():
    """A better str(_get_keycodes()) method"""
    keycodes = _get_keycodes()
    initial_code, codes = keycodes[0], keycodes[1:]
    initial_char = chr(initial_code)
    if initial_code == 27:
        initial_char = '\\e'
    elif not ascii.isgraph(initial_char):
        initial_char = '\\x%x' % initial_code
    chars = ''.join([chr(c) for c in codes])
    return ''.join((initial_char, chars))


def yield_asciis_old():
    """This one is deprecated"""
    k = get_ascii()
    while True:
        yield k


def ask_user(prompt, default=None):
    prompt_default = str('[%s]' % default) if default else ''
    print('%s %s? ' % (prompt, prompt_default), end='')
    result = getch().lower().strip()
    print()
    return result or default



#! /usr/bin/env python

import argparse
import os
import sys
from functools import partial
from itertools import chain

_exit_ok = os.EX_OK
_exit_fail = not _exit_ok

from pysyte.debuggers import DebugExit


def extract_strings(names, name):
    result = {}
    try:
        value = names[name]
    except KeyError:
        return []
    if isinstance(value, str):
        return [value]
    try:
        return [v for v in value if isinstance(v, str)]
    except TypeError:
        return []


def parser(description=None, usage=None):
    """Parse out command line arguments"""

    class ArgNamespace(object):
        def __init__(self, result):
            self._result = result

        def __getattr__(self, name):
            try:
                super(ArgNamespace, self).__getattr__(name)
            except AttributeError:
                return getattr(self._result, name)

        def arg_strings(self, name=None):
            if not self._result:
                return None
            return extract_strings(self._result.__dict__, name)

    class ArgParser(object):

        def __init__(self, argument_parser):
            self.parser = argument_parser
            self.parse = self.parse_args
            self._parsed = None

            self.arg = self.parser.add_argument
            self.opt = self.true = partial(self.arg, action='store_true')
            self.int = partial(self.arg, type=int)

        def parse_args(self):
            return ArgNamespace(self.parser.parse_args())

        def sub(self, name, help=''):
            self.arg(name, metavar=name, help=help, type=str)

        def args(self, name, help='', many='*'):
            self.arg(name, metavar=name, help=help, type=str, nargs=many)

        def answer(self, answers=None):
            yes_no = {
                'n': ('-n', '--no', 'Answer no'),
                'y': ('-y', '--yes', 'Answer yes'),
            }
            options = yes_no
            options.update(answers)
            [self.true(options[o]) for o in options]

    lines = description.splitlines()
    if not usage:
        if len(lines) > 1 and not lines[1]:
            usage, epilog = lines[0], '\n\n'.join(lines[2:])
    else:
        usage = description = lines[0]
        epilog = None
    return ArgParser(
        argparse.ArgumentParser(usage=description))


def arg_strings(parsed_args, name=None):
    """A list of all strings for the named arg"""
    name = name or 'arg_strings'
    value = getattr(parsed_args, name, [])
    if isinstance(value, str):
        return [value]
    try:
        return [v for v in value if isinstance(v, str)]
    except TypeError:
        return []


def main_parser(outer_main, parser):

    def main():
        outer_main(parser().parse_args())

    call_main(main)


def call_main(main):

    def unhandled(_e, message):
        """Placeholder: write to log / Slack

        e.g.
        type_, value, traceback = sys.exc_info()
        thing = format_exception(type_, value, traceback)
        """
        print(message, file=sys.stderr)
        raise  # pylint: disable=misplaced-bare-raise

    try:
        return _exit_ok if main() else _exit_fail
    except SystemExit as e:
        return e.code
    except KeyboardInterrupt as e:
        ctrl_c = 3
        return ctrl_c
    except DebugExit:
        return _exit_fail
    except AssertionError as e:
        return unhandled(e, 'Failed Assertion')
    except Exception as e:  # pylint: disable=broad-except
        return unhandled(e, 'Unhandled Exception')


def run_main(main, add_args):
    def inner_main():
        args = parser_.parse_args()
        return main(args)

    parser_ = parser(add_args.__module__.__doc__)
    add_args(parser_)
    sys.exit(call_main(inner_main))



# pylint: disable=unused-import

try:
    from commands import getstatusoutput
    from commands import getoutput
except ImportError:
    import os
    import subprocess

    from subprocess import getoutput

    def getstatusoutput(command):
        status, output = subprocess.getstatusoutput(command)
        if os.name != 'nt':
            # convert status to be interpreted according to the wait() rules
            status = status << 8
        return status, output



def init(self):

    def get():
        return self

    def set(value):
        nonlocal self
        self = value
        return self

    setattr(self, 'set', set)
    setattr(self, 'get', get)
    self.get = get
    return self



from pysyte.bash import cmnds


class NoTerminalAvailable(NotImplementedError):
    pass


def _tput(tput_command):
    command = 'tput %s' % tput_command
    status, output = cmnds.getstatusoutput(command)
    if status:
        if 'No value for $TERM' in output:
            raise NoTerminalAvailable(output)
        raise NotImplementedError(output)
    if not output:
        return 0
    try:
        return int(output) - 1
    except TypeError:
        return 0


def screen_width():
    return _tput('cols')


def screen_height():
    return _tput('lines')



"""Convert between some numbers known to represent colours

Methods handle two known representations:
    1) hex RGB
        Methods are provided for strings like "0xFF0000" (for red)
            and the equivalent numbers
        See http://en.wikipedia.org/wiki/Web_colors#Shorthand_hexadecimal_form
    2) ANSI
        Convert RGB to ANSI values
        See http://en.wikipedia.org/wiki/ANSI_escape_code#Colors
"""


import re


from pysyte.colours import colour_names


def integer_to_ansi(integer):
    if integer < 16:
        return integer

    def rgb_to_decimal(i):
        i -= 0x10
        if i < 0:
            return 0
        return i // 40

    def grey_shade(red, green, blue):
        for i in red, green, blue:
            if (i - 8) % 10:
                return False
        return 232 + ((red - 8) // 10)

    red = (integer & 0xff0000) >> 16
    green = (integer & 0x00ff00) >> 8
    blue = integer & 0x0000ff
    result = grey_shade(red, green, blue)
    if result:
        return result
    red = rgb_to_decimal(red)
    green = rgb_to_decimal(green)
    blue = rgb_to_decimal(blue)
    return (red * 36) + (green * 6) + blue + 16


def ansi_to_red_green_blue(ansi):
    a = ansi
    assert 231 >= a >= 16
    red, green, blue = ((a - 16) / 36) % 6, ((a - 16) / 6) % 6, (a - 16) % 6

    def decimal_to_rgb(i):
        if not i:
            return 0
        return int((i * 40) + 55)
    return decimal_to_rgb(red), decimal_to_rgb(green), decimal_to_rgb(blue)


def red_green_blue_to_int(red, green, blue):
    assert 0 <= red < 512
    assert 0 <= blue < 512
    assert 0 <= green < 512, green
    return (red << 16) + (green << 8) + blue


def _hex_regexp():
    return re.compile('(([0-9a-f]{6})|([0-9a-f]{3}))', re.IGNORECASE)


def _extract_html_hex(string):
    """Get the first 3 or 6 hex digits in the string"""
    try:
        hex_string = string and _hex_regexp().search(string).group(0) or ''
    except AttributeError:
        return None
    if len(hex_string) == 3:
        hex_string = hex_string[0] * 2 + hex_string[1] * 2 + hex_string[2] * 2
    return hex_string


def html_to_int(string):
    """Use the first 3 or 6 hex digits in the string as a html colour picker"""
    hex_string = _extract_html_hex(string)
    if hex_string is None:
        return None
    if not hex_string:
        return 0
    return int(hex_string, 16)


def html_to_html(string):
    digits = string.lstrip('#').upper()
    if len(digits) == 3:
        return ''.join([d * 2 for d in digits])
    return digits


def hashed_html(string):
    return '#%s' % html_to_html(string)


def html_to_red_green_blue(string):
    string = _extract_html_hex(string)
    if string is None:
        return None, None, None
    return int(string[:2], 16), int(string[2:4], 16), int(string[4:], 16)


def integer_to_html(integer):
    return '#%06X' % integer


def html_to_ansi(string):
    i = html_to_small_ansi(string)
    if i is not None:
        return i
    i = html_to_int(string)
    return integer_to_ansi(i)


def ansi_to_int(ansi):
    if ansi <= 16:
        return ansi
    elif ansi > 231:
        i = ansi - 232
        r = g = b = (i * 10) + 8
    else:
        r, g, b = ansi_to_red_green_blue(ansi)
    return red_green_blue_to_int(r, g, b)


def small_values():
    return [
        (0x00, '#000000'),
        (0x01, '#800000'),
        (0x02, '#008000'),
        (0x03, '#808000'),
        (0x04, '#000080'),
        (0x05, '#800080'),
        (0x06, '#008080'),
        (0x07, '#C0C0C0'),
        (0x08, '#808080'),
        (0x09, '#FF0000'),
        (0x0A, '#00FF00'),
        (0x0B, '#FFFF00'),
        (0x0C, '#0000FF'),
        (0x0D, '#FF00FF'),
        (0x0E, '#00FFFF'),
        (0x0F, '#000000'),
    ]


def small_integers():
    return [integer for integer, _html in small_values()]


def small_ansi_to_html(i):
    for ansi, html in small_values():
        if i == ansi:
            return html
    return None


def html_to_small_ansi(string):
    hashed = hashed_html(string)
    for ansi, html in small_values():
        if html == hashed:
            return ansi
    return None


def ansi_to_html(ansi):
    if ansi < 16:
        return small_ansi_to_html(ansi)
    i = ansi_to_int(ansi)
    return integer_to_html(i)


def name_to_int(name):
    """Get a number for that colour name

    if not a name, then not a number
    """
    if not name:
        return float('nan')
    lower = name.lower()
    cga_names = {s: i for i, s in enumerate(colour_names.cga())}
    return cga_names.get(lower) or html_to_small_ansi(lower)



"""This module provides methods to cover many git commands"""
import logging
import os
import re
import shutil
from contextlib import contextmanager
from functools import partial

from subprocess import getstatusoutput

_working_dirs = ['.']


logger = logging.getLogger('tools')


class GitException(RuntimeError):
    """This class is the base for git exceptions

    For many git exceptions there are known workarounds
        E.g. if a commit didn't work, then add some files first!

    But that error (nothing to commit) can arise in many contexts
        eg as a "Git Commit Exception"
        or as a "Git Merge Exception"
        and those 2 exception classes might not be derived from each other
    So the merge exception cannot rely on inheriting an "add_files_handler()"
        It cannot check bases classes, so must check derived classes

    This metaclass allows any class derived from it to insist
        that any further-derived classes provide some handlers
    """

    required_handlers = []

    def __new__(cls, name, bases, body):
        """Derived classes must have all required_handlers"""
        if 'required_handlers' in body:
            missing = set(cls.required_handlers) - set(body.keys())
            if missing:
                raise TypeError(
                    'Missing handler%s: %s' % (
                        len(missing) == 1 and '' or 's',
                        '\n'.join(
                            '%s.%s()' % (cls.__class__.__name__, m)
                            for m in missing)))
        return super(GitException, cls).__new__(cls, name, bases, body)


class GitError(GitException):
    def __init__(self, command, status_, output):
        super(GitError, self).__init__(
            '\nPWD: %r\nCommand: %r\nstderr:\n%r' % (
                os.getcwd(), command, output))
        self.command = command
        self.status = status_
        self.output = output

    def inits(self):
        return self.command, self.status, self.output


# pylint: disable=too-many-ancestors
class UnknownRevision(GitError):
    pass


class NoRemoteRef(GitError):
    pass


class NoSuchCommit(GitError):
    pass


class ExistingReference(GitError):
    pass


class ExistingBranch(GitError):
    pass


class EmptyCommit(GitError):
    pass


def resolve_conflicts(in_lines):
    lines = []
    resolved = set()
    conflicts = set()
    for line in in_lines:
        if line.startswith('Resolved'):
            name = line.split("'")[1]
            resolved.add(name)
            add(name)
            if name in conflicts:
                conflicts.remove(name)
            continue
        lsw = line.startswith
        if lsw('CONFLICT') or lsw('Recorded preimage'):
            if lsw('CONFLICT'):
                name = line.split()[-1]
            else:
                name = line.split()[-1].strip("'")
            if name in resolved or name in conflicts:
                continue
            conflicts.add(name)
        lines.append(line)
    return '\n'.join(lines), resolved, conflicts


class Resolver(GitError):
    required_handlers = ['resolve_conflicts']

    def __init__(self, command, status_, output):
        super(Resolver, self).__init__(command, status_, output)
        self.output, self.resolved, self.conflicts = self.resolve_conflicts()

    def resolve_conflicts(self):
        """Remove files that git says are resolved from the conflicts"""
        return resolve_conflicts(self.output.splitlines())


# importers can assign to this to change the Exception raised
# new Exception should derive from Resolver
class ResolveError(Resolver):
    pass


@contextmanager
def use_resolver(resolver):
    # pylint: disable=global-variable-undefined
    global ResolveError
    old = ResolveError
    assert issubclass(resolver, GitError)
    try:
        ResolveError = resolver
        yield resolver
    finally:
        ResolveError = old


def root_from(path):
    p = path if os.path.isdir(path) else os.path.dirname(path)
    while p[:2] >= '/A':
        p_git = os.path.join(p, '.git')
        if os.path.isdir(p_git):
            return p
        p = os.path.dirname(p)
    raise ValueError('No .git above %s' % path)


@contextmanager
def pushd(path):
    _working_dirs.insert(0, root_from(path))
    yield _working_dirs[0]
    del _working_dirs[0]


def cd(path):
    """Change the directory that git will use for subsequent commands

    method is named for consistency with similar in shell.py
    Actually it just adds the "-C" option to git commands

    If this method is not called then commands default to current directory
        i.e. the global _working_dirs, above, defaults to '.'
    """
    _working_dirs[0] = root_from(path)


def run(sub_command, quiet=False, no_edit=False, no_verify=False):
    """Run a git command

    Prefix that sub_command with "git "
        then run the command in shell

    If quiet if True then do not log results
    If no_edit is True then prefix the git command with "GIT_EDITOR=true"
        (which does not wait on user's editor)

    If the command gives a non-zero status, raise a GitError exception
    """
    if _working_dirs[0] != '.':
        git_command = 'git -C "%s"' % _working_dirs[0]
    else:
        git_command = 'git'
    edit = 'GIT_EDITOR=true' if no_edit else ''
    verify = 'GIT_SSL_NO_VERIFY=true' if no_verify else ''
    command = '%s %s %s %s' % (verify, edit, git_command, sub_command)
    if not quiet:
        logger.info('$ %s', command)
    status_, output = getstatusoutput(command)
    if status_:
        if quiet:
            logger.info('$ %s', command)
        logger.error('\n%s', output)
        if 'unknown revision' in output:
            raise UnknownRevision(command, status_, output)
        elif 'remote ref does not exist' in output:
            raise NoRemoteRef(command, status_, output)
        elif 'no such commit' in output:
            raise NoSuchCommit(command, status_, output)
        elif 'reference already exists' in output:
            raise ExistingReference(command, status_, output)
        elif re.search('Resolved|CONFLICT|Recorded preimage', output):
            raise ResolveError(command, status_, output)
        elif re.search('branch named.*already exists', output):
            raise ExistingBranch(command, status, output)
        raise GitError(command, status_, output)
    elif output and not quiet:
        logger.info('\n%s', output)
    return output


def log(args, number=None, oneline=False, quiet=False):
    """Run a "git log ..." command, and return stdout

    args is anything which can be added after a normal "git log ..."
        it can be blank
    number, if true-ish, will be added as a "-n" option
    oneline, if true-ish, will add the "--oneline" option
    """
    options = ' '.join([
        number and str('-n %s' % number) or '',
        oneline and '--oneline' or ''
    ])
    try:
        return run('log %s %s' % (options, args), quiet=quiet)
    except UnknownRevision:
        return ''


def rev_parse(options, *args, **kwargs):
    """Run 'git rev-parse' with those options"""
    return run('rev-parse %s' % options, *args, **kwargs)


def root():
    """Get the root directory of the repo

    i.e. this directory (or one above) which contains the ".git" directory
    """
    return os.path.dirname(rev_parse('--absolute-git-dir'))


def branch(options=False, *args, **kwargs):
    """Run "$ git branch" with those options

    If not options then return name of the branch currently checked out
    """
    return (options
            and run('branch %s' % options, *args, **kwargs)
            or rev_parse('--abbrev-ref HEAD', *args, **kwargs))


def branches(remotes=False):
    """Return a list of all local branches in the repo

    If remotes is true then also include remote branches

    Note: the normal '*' indicator for current branch is removed
        this method just gives a list of branch names
    Use branch() method to determine the current branch
    """
    stdout = branch('--list %s' % (remotes and '-a' or ''), quiet=True)
    return [_.lstrip('*').strip() for _ in stdout.splitlines()]


def branches_containing(commit):
    """Return a list of branches conatining that commit"""
    lines = run('branch --contains %s' % commit).splitlines()
    return [l.lstrip('* ') for l in lines]


def hide(item):
    """Hide that item (branch or tag), i.e. remove it from origin"""
    try:
        return push('origin --delete', item)
    except NoRemoteRef:
        pass


def unstaged_files():
    """A list of all filenames which have changes, but not staged for commit"""
    return diff_files('', '')


def conflicted(path_to_file):
    """Whether there are any conflict markers in that file"""
    for line in open(path_to_file, 'r'):
        for marker in '>="<':
            if line.startswith(marker * 8):
                return True
    return False


def add(path=None, force=False, quiet=False):
    """Add that path to git's staging area (default current dir)

    so that it will be included in next commit
    """
    option = '-f' if force else ''
    return run('add %s %s' % (option, path) or '.', quiet=quiet)


def push(qualifiers='', refspec=None):
    """Push the local refspec to remote repository

    If refspec is left as None, then push current branch
    If any qualifiers are given, then they will be added to the push command
    """
    run('push %s %s' % (qualifiers, refspec or ''))


def fetch_all():
    """Fetch all (new) branches and tags from remote"""
    run('fetch --all')


def grup():
    """Prune all local branches and tags removed from remote

    method name is an acronym of the command
    """
    run('remote update origin --prune')


def config(key, value, local=True):
    """Set that config key to that value

    Unless local is set to False: only change local config
    """
    option = local and '--local' or ''
    run('config %s "%s" "%s"' % (option, key, value))


def clean(full=False):
    options = '-d -f -f -x' if full else ''
    run('clean %s' % options)
    run('gc --quiet')
    if full:
        run('reset HEAD .')
        checkout('.')

def clone(url, path=None, remove=True):
    """Clone a local repo from that URL to that path

    If path is not given, then use the git default: same as repo name
    If path is given and remove is True
        then the path is removed before cloning

    Because this is run from a script it is assumed that user should be Admin
        so set config user values for the GitLab Admin
    """
    clean = True
    if path and os.path.isdir(path):
        if not remove:
            clean = False
        else:
            shutil.rmtree(path)
    if clean:
        stdout = run('clone %s %s' % (url, path or ''))
        into = stdout.splitlines()[0].split("'")[1]
        path_to_clone = os.path.realpath(into)
    else:
        path_to_clone = path
    old_dir = _working_dirs[0]
    _working_dirs[0] = path_to_clone
    config('user.name', 'Release Script')
    config('user.email', 'gitlab@wwts.com')
    _working_dirs[0] = old_dir
    return path_to_clone


def diff_files(branch_1, branch_2):
    return run('diff --name-only %s %s' % (branch_1, branch_2)).splitlines()


def status(short=False):
    option = '--short' if short else '--long'
    return run('status %s' % option)


def needs_abort():
    """A command to abort an operation in progress

    For example a merge, cherry-pick or rebase
    If one of these operations has left the repo conflicted
        then give a command to abandon the operation
    """
    for line in status().splitlines():
        if '--abort' in line:
            for part in line.split('"'):
                if '--abort' in part:
                    return part
        elif 'All conflicts fixed but you are still merging' in line:
            return 'git merge --abort'
        elif 'You have unmerged paths.' in line:
            return 'git merge --abort'
        elif 'all conflicts fixed: run "git rebase --continue"' in line:
            return 'git rebase --abort'
    return None


@contextmanager
def very_clean():
    """Do a full git clean before and after"""
    abandon_fully()
    try:
        yield
    finally:
        abandon_fully()


def abandon_fully():
    abandon_operation()
    clean(full=True)

def abandon_operation():
    """Abandon any current operaton which needs it"""
    command = needs_abort()
    if not command:
        return False
    return run(command.replace('git', '', 1))


def show_branches(branch1, branch2):
    """Runs git show-branch between the 2 branches, parse result"""

    def find_column(string):
        """Find first non space line in the prefix"""
        result = 0
        for c in string:
            if c == ' ':
                result += 1
        return result

    def parse_show_line(string):
        """Parse a typical line from git show-branch

        >>> parse_show_line('+  [master^2] TOOLS-122 Add bashrc')
        '+ ', 'master^2', 'TOOLS-122 Add bashrc'
        """
        regexp = re.compile(r'(.*)\[(.*)] (.*)')
        match = regexp.match(string)
        if not match:
            return None
        prefix, commit, comment = match.groups()
        return find_column(prefix), commit, comment

    log = run('show-branch --sha1-name "%s" "%s"' % (branch1, branch2))
    lines = iter(log.splitlines())
    line = lines.next()
    branches = {}
    while line != '--':
        column, branch, comment = parse_show_line(line)
        branches[column] = [branch]
        line = lines.next()
    line = lines.next()
    for line in lines:
        column, commit, comment = parse_show_line(line)
        branches[column].append((commit, comment))
    return branches


# pylint: disable=redefined-outer-name
# From here down args may redefine methods
# This arises from trying to match git commands
#   where one command may appear as an option in another command
# For example
#   The word "commit" is a "command" in "$ git commit ..."
#   But it is an "option" in "$ git tag ... commit"
#
# Callers should not have to use mangled names for args / options

# Define aliases for those defined above and used below:
current_branch = branch


def latest_commit(branch=None, path=None):
    """Last commit message on given (or current) branch

    If a path is given, then get latest for that path only
    """
    options = '--no-abbrev-commit %s %s' % (
        branch or '',
        path and str('-- %s' % path) or '')
    result = log(options, 1, True, True)
    return result and result.split(' ', 1) or ('', '')


def commits_with_message(message):
    """All commits with that message (in current branch)"""
    output = log("--grep '%s'" % message, oneline=True, quiet=True)
    lines = output.splitlines()
    return [l.split(' ', 1)[0] for l in lines]


def sha1(branch=None, path=None):
    """SHA1 of the last commit on given (or current) branch

    If a path is given, then get latest for that path only
    """
    return latest_commit(branch, path)[0]


def checkout(branch, quiet=False, as_path=False):
    """Check out that branch

    Defaults to a quiet checkout, giving no stdout
        if stdout it wanted, call with quiet = False

    Defaults to checking out branches
        If as_path is true, then treat "branch" like a file, i.e.
        $ git checkout -- branch

    All errors will pass silently, just returning False
        except any messages about "you need to resolve your current index"
        These indicate that the repository is not in a normal state
            and action by a user is usually needed to resolve that
        So the exception is allowed to rise
            probably stopping the script
    """
    try:
        if as_path:
            branch = '-- %s' % branch
        run('checkout %s %s' % (quiet and '-q' or '', branch))
        return True
    except GitError as e:
        if 'need to resolve your current index' in e.output:
            raise
        return False


checkout_path = partial(checkout, as_path=True)


def diff(options=None, branch='master', path=None):
    suffix = path and str('-- %s' % path) or ''
    return run('diff %s %s %s' % (options, branch, suffix))


def same_diffs(commit1, commit2):
    """Whether those 2 commits have the same change

    Run "git range-diff" against 2 commit ranges:
        1. parent of commit1 to commit1
        1. parent of commit2 to commit2

    If the two produce the same diff then git will only show header lines
        Header lines are prefixed like "1: ...."
    Otherwise there will be lines after the header lines
        (and those lines will show the difference)
    """
    def range_one(commit):
        """A git commit "range" to include only one commit"""
        return '%s^..%s' % (commit, commit)

    output = run('range-diff %s %s' % (range_one(commit1), range_one(commit2)))
    lines = output.splitlines()
    for i, line in enumerate(lines, 1):
        if not line.startswith('%s: ' % i):
            break
    return not lines[i:]


def checkout_log(branch, number=15, quiet=False, as_path=False):
    if checkout(branch, quiet, as_path):
        return log('', number=number, oneline=True, quiet=quiet).splitlines()
    return []


def renew_local_branch_from_remote(branch, start_point):
    checkout(start_point)
    pulled = pull()
    renew_local_branch(branch, start_point)
    return pulled


def renew_local_branch(branch, start_point, remote=False):
    """Make a new local branch from that start_point

    start_point is a git "commit-ish", e.g branch, tag, commit

    If a local branch already exists it is removed
    If remote is true then push the new branch to origin
    """
    if branch in branches():
        checkout(start_point)
        delete(branch, force=True, remote=remote)
    result = new_local_branch(branch, start_point)
    if remote:
        publish(branch)
    return result


def pull_branch(branch):
    checkout(branch)
    pull()
    return up_to_date()

def up_to_date():
    return bool(re.search('up.to.date', status()))

def new_local_branch(branch, start_point):
    """Make a new local branch from that start_point

    start_point is a git "commit-ish", e.g branch, tag, commit
    """
    return run('checkout -b %s %s' % (branch, start_point))


def publish(branch, full_force=False):
    """Publish that branch, i.e. push it to origin"""
    checkout(branch)
    try:
        push('--force --set-upstream origin', branch)
    except ExistingReference:
        if full_force:
            push('origin --delete', branch)
            push('--force --set-upstream origin', branch)



def delete(branch, force=False, remote=False):
    option = '-D' if force else '-d'
    try:
        result = run('branch %s %s' % (option, branch))
        pass
    except GitError as e:
        if 'checked out at' not in e.output:
            raise
        if branch == 'master':
            raise
        checkout('master')
        result = run('branch %s %s' % (option, branch))
    except Exception as e:
        pass
    if remote:
        return hide(branch)
    return result


def commit(message, add=False, quiet=False):
    """Commit with that message and return the SHA1 of the commit

    If add is truish then "$ git add ." first
    """
    if add:
        run('add .')
    try:
        stdout = run('commit -m %r' % str(message), quiet=quiet)
    except GitError as e:
        s = str(e)
        if 'nothing to commit' in s or 'no changes added to commit' in s:
            raise EmptyCommit(*e.inits())
        raise
    return re.split(r'[ \]]', stdout.splitlines()[0])[1]


def pull(rebase=True, refspec=None):
    """Pull refspec from remote repository to local

    If refspec is left as None, then pull current branch
    The '--rebase' option is used unless rebase is set to false
    """
    options = rebase and '--rebase' or ''
    output = run('pull %s %s' % (options, refspec or ''))
    return not re.search('up.to.date', output)


def tag(name, remote=False, commit=None):
    """Add a local tag with that name at that commit

    If no commit is given, at current commit on cuurent branch
    If remote is true, also push the tag to origin
    """
    command = 'tag %s %s' % (name, commit or '')
    result = run(command)
    if remote:
        push('origin %s' % name)
    return result


def tags():
    return run('tag', quiet=True).splitlines()


def in_tag(name):
    for tag in tags():
        if name in tag:
            return True
    return False


def is_tag(name):
    return name in tags()


def re_tag(name, remote=False, commit=None):
    """Add a local tag with that name at that commit

    If no commit is given, at current commit on cuurent branch
    If remote is true, also push the tag to origin

    If tag already exists, delete it, then re-make it
    """
    if remote:
        tags = run('ls-remote --tags', quiet=True).splitlines()
        if name in tags:
            hide(name)
    if is_tag(name):
        run('tag --delete %s' % name)
    return tag(name, remote, commit)


def merge(branch, fast_forward=False, commit=True, options=None):
    """Merge that branch into current branch

    fast_forward and commit add the eponymous options to the command
    """
    more_options = [
        fast_forward and '--ff' or '--no-ff',
        commit and '--commit' or '--no-commit',
    ] + (options and options or [])
    option_string = ' '.join(more_options)
    all_branches = branches()
    assert branch in all_branches, 'Missing branch: %s' % branch
    command = 'merge %s %s' % (option_string, branch)
    return run(command)


@contextmanager
def git_continuer(method, *args, **kwargs):
    try:
        yield
    except Resolver as resolved:
        while True:
            if resolved.conflicts:
                raise
            try:
                method(*args, **kwargs)
            except Resolver as resolved:
                pass
            else:
                return


def cherry_pick(pick):
    """Cherry pick that arg into current branch

    pick would normally just be SHA1 of a commit,
        but can include command options, e.g.:
            cherry_pick('ab12cd34 -X theirs')
    """
    with git_continuer(run, 'cherry-pick --continue', no_edit=True):
        return run('cherry-pick -x --allow-empty %s' % pick)


def rebase(upstream, branch=None):
    """Rebase branch onto upstream

    If branch is empty, use current branch

    """
    rebase_branch = branch and branch or current_branch()
    with git_continuer(run, 'rebase --continue', no_edit=True):
        stdout = run('rebase %s %s' % (upstream, rebase_branch))
        return 'Applying' in stdout


def merge_resolve(branch, fast_forward=False, commit=True, options=None):
    """Merge that branch into current branch

    If git raises an error, and the error is for conflicts
        then read the error message looking for any resolved conflicts
        Remove the resolved conflicts from the error message
    If all conflicts were resolved
        then remove the error
    """
    def get_resolution(resolved=None):
        return resolved

    with git_continuer(get_resolution):
        resolved = merge(branch, fast_forward, commit, options)
        if commit and resolved.resolved:
            # Then the resolved files were added, better commit them
            commit("Merge branch '%s' into '%s'" % (branch, current_branch()))


def split_conflict(conflict):
    position = before = []
    during = []
    after = []
    previous_conflict = False
    with open(conflict) as stream:
        for line in stream:
            markers = [
                '<<<<<<<',
                '|||||||',
                '=======',
                '>>>>>>>',
            ]
            for marker in markers:
                if line.startswith(marker):
                    if marker == '<<<<<<<':
                        position = during
                        # Assuming only one conflict in a version file
                        assert previous_conflict is False
                        previous_conflict = True
                    elif marker == '>>>>>>>':
                        position = after
                    break
            else:
                position.append(line)
    return before, during, after



import os
from pysyte.colours import ansi_escapes
from pysyte.colours import colour_numbers
from pysyte.colours.texts import colour_initials


def test():
    return ansi_escapes.foreground_string(
        'hello', colour_numbers.html_to_int('0x00FF00'))


def main():
    for b in range(8):
        print(ansi_escapes.background(b), end='')
        for f in range(8):
            print('%s%s%s%s%s' % (
                ansi_escapes.foreground(f),
                ansi_escapes.bold(),
                'hello ',
                ansi_escapes.no_bold(),
                'world'), end='')
        print(ansi_escapes.no_colour())

    term = os.environ.get('TERM', '')
    if '256' in term or term == 'linux':
        for foregound in ['8b4513', '00ff7f', 'ff1493']:
            print(ansi_escapes.foreground_string(
                'hello', colour_numbers.html_to_int(foregound)))
        print(ansi_escapes.no_colour())
    print(colour_initials(['Hello ', 'World'], 'red').colour('!', 'green'))


if __name__ == '__main__':
    main()




"""Some methods to help with the handling of dictionaries"""


from collections import defaultdict

def get_caselessly(dictionary, sought):
    """Find the sought key in the given dictionary regardless of case

    >>> things = {'Fred' : 9}
    >>> print(get_caselessly(things, 'fred'))
    9
    """
    try:
        return dictionary[sought]
    except KeyError:
        caseless_keys = {k.lower(): k for k in dictionary.keys()}
        real_key = caseless_keys[sought.lower()]  # allow any KeyError here
        return dictionary[real_key]


def swap_dictionary(dictionary):
    """Swap keys for values in the given dictionary

    >>> swap_dictionary({'one': 1})[1] == 'one'
    True
    """
    if dictionary is None:
        return None
    return {v: k for k, v in dictionary.items()}


def append_value(dictionary, key, item):
    """Append those items to the values for that key"""
    items = dictionary.get(key, [])
    items.append(item)
    dictionary[key] = items


def extend_values(dictionary, key, items):
    """Extend the values for that key with the items"""
    values = dictionary.get(key, [])
    try:
        values.extend(items)
    except TypeError:
        raise TypeError('Expected a list, got: %r' % items)
    dictionary[key] = values


def group_list(items):
    """Items in the list are grouped into a dictionary

    Items should be a list of (key, value) pairs

    >>> grouped = group_list([(1,0), (2,0), (1,1)])
    >>> grouped[1] == [0,1]
    True
    """
    groups = defaultdict(list)
    for key, value in items:
        groups[key].append(value)
    return groups


def group_list_by(items, key_from_item):
    """Items in the list are grouped into a dictionary

    key_from_item is a method to get the key from the item

    >>> items = [(1,9), (2,8), (1,7)]
    >>> key_from_item = lambda x: 'a' if x[0] == 1 else 'b'
    >>> grouped = group_list_by(items, key_from_item)
    >>> grouped['a'] == [(1, 9), (1, 7)]
    True
    """
    groups = defaultdict(list)
    for item in items:
        key = key_from_item(item)
        groups[key].append(item)
    return groups


class DefaultDict(defaultdict):
    """A default dict with improved repr"""
    def __repr__(self):
        from pprint import pformat
        return '<%s\n    %s\n>' % (
            self.__class__.__name__,
            pformat(dict(self), indent=4))


class LazyDefaultDict(DefaultDict):
    """A defaultdict which waits for key access to provide default

    Same as a defaultdict, but the initiasing method takes a key
        and provides a value
    """
    def __init__(self, method):
        self.method = method
        super(LazyDefaultDict, self).__init__(None)

    def __missing__(self, key):
        result = self.method(key)
        self[key] = result
        return result



"""Literal values generally useful"""

# pylint: disable=no-init


class Literals(object):
    pass


class Empty(Literals):
    string = ''
    integer = 0
    real = 0.0


class Punctuation(Literals):
    space = ' '
    comma = ','
    full_stop = dot = period = '.'
    underline = '_'
    forward_slash = '/'
    backslash = '\\'
    colon = ':'
    asterisk = '*'
    minus = '-'
    plus = '+'
    question = '?'
    double_quote = '"'
    apostrophe = single_quote = "'"
    zero = '0'
    at = '@'
    ampersand = '&'
    equal = equals = '='


class Digits(Literals):
    false = zero = '0'
    true = one = '1'
    two = '2'
    three = '3'
    four = '4'
    five = '5'
    six = '6'
    seven = '7'
    eight = '8'
    nine = '9'


class Numbers(Digits):
    minus_one = '-1'
    ten = '10'
    eleven = '11'
    twelve = '12'
    thirteen = '13'


class Python(Literals):
    name = directory = command = package = inline = 'python'
    extension = 'py'
    ext = '.py'


class AnsiColourSequences(Literals):
    """Colour sequences are derived from the table at

    http://en.wikipedia.org/wiki/ANSI_escape_sequences#Colors"""
    light_colour = "\033[1;%sm"
    light_black = light_colour % 30
    light_red = light_colour % 31
    light_green = light_colour % 32
    yellow = light_colour % 33
    light_blue = light_colour % 34
    light_magenta = light_colour % 35
    light_cyan = light_colour % 36
    white = light_colour % 37
    dark_colour = "\033[0;%sm"
    black = light_colour % 30
    red = dark_colour % 31
    green = dark_colour % 32
    brown = dark_colour % 33
    blue = dark_colour % 34
    magenta = dark_colour % 35
    cyan = dark_colour % 36
    grey = dark_colour % 37
    off = "\033[0m"

    @classmethod
    def highlighted(cls, colour, text):
        return '%s%s%s' % (colour, text, cls.off)



"""Methods to provide X11 names for colours

If there is an "rgb.txt" file in a known directory get names from that
    otherwise fall back on the standard 8 names (BRGBCYMW)

The rgb.txt file is expected to contain X11 colour names
    so the "known directory" is <X11root>/lib/X11/
        Some variations on that official path are also tried
    For a non-standard path, call use_rgb_txt()
    See http://en.wikipedia.org/wiki/X11_color_names
"""


import os
import re


def _rgb_txt_directories():
    return [
        '/usr/lib/X11',
        '/usr/X11/share/X11',
        '.',
        os.path.dirname(__file__),
    ]


def use_rgb_txt(path_to_rgb_txt):
    def new_method():
        return [path_to_rgb_txt]

    if not os.path.isfile(path_to_rgb_txt):
        return False
    # Yes, we are "Redefining name '_rgb_txt_directories' from outer scope"
    # pylint: disable-msg=W0621
    _rgb_txt_directories = new_method
    if hasattr(_rgb_txt_names_and_numbers, 'result'):
        delattr(_rgb_txt_names_and_numbers, 'result')
    return True


def first_rgb_file(rgb_txt_paths):
    for path in rgb_txt_paths():
        path_to_txt_file = os.path.join(path, 'rgb.txt')
        if os.path.isfile(path_to_txt_file):
            return path_to_txt_file
    return None


def _rgb_txt_line(string):
    regexp = re.compile(r'([ 0-9][ 0-9][ 0-9])\s+([ 0-9][ 0-9][ 0-9])\s+([ 0-9][ 0-9][ 0-9])\s+([a-zA-Z0-9 ]+)\s*')
    match = regexp.match(string)
    if not match:
        return '', None
    red, green, blue, name = match.groups()
    return name.strip(), (int(red), int(green), int(blue))


def _rgb_txt_names_and_numbers(path_to_file):
    if not path_to_file:
        return []
    if not hasattr(_rgb_txt_names_and_numbers, 'result'):
        rgb_lines = [_rgb_txt_line(l) for l in open(path_to_file)]
        _rgb_txt_names_and_numbers.result = [(name, values) for name, values in rgb_lines if name]
    return _rgb_txt_names_and_numbers.result


def _local_rgb_txt_names_and_numbers():
    path_to_rgb_txt = first_rgb_file(_rgb_txt_directories)
    return _rgb_txt_names_and_numbers(path_to_rgb_txt)


def known_colour_names_and_numbers():
    known_names = 'black red green yellow blue magenta cyan white'
    dark_names_and_numbers = [(name, number) for number, name  in enumerate(known_names.split())]
    light_names_and_numbers = [('light %s' % name, number + 8) for name, number in dark_names_and_numbers]
    return dark_names_and_numbers + light_names_and_numbers


def names():
    names_and_numbers = _local_rgb_txt_names_and_numbers()
    if not names_and_numbers:
        names_and_numbers = known_colour_names_and_numbers()
    return dict(names_and_numbers)





"""A module to provide decorators which change methods"""

from six import StringIO


def _represent_arguments(*arguments, **keyword_arguments):
    """Represent the aruments in a form suitable as a key (hashable)

    And which will be recognisable to user in error messages
    >>> print(_represent_arguments([1, 2], **{'fred':'here'}))
    [1, 2], fred='here'
    """
    argument_strings = [repr(a) for a in arguments]
    keyword_strings = [
        '='.join((k, repr(v))) for k, v in keyword_arguments.items()]
    return ', '.join(argument_strings + keyword_strings)


def memoize(method):
    """A new method which acts like the given method but memoizes arguments

    See https://en.wikipedia.org/wiki/Memoization for the general idea
    >>> @memoize
    ... def test(arg):
    ...     print('called')
    ...     return arg + 1
    >>> test(1)
    called
    2
    >>> test(2)
    called
    3
    >>> test(1)
    2

    The returned method also has an attached method "invalidate"
        which removes given values from the cache
        Or empties the cache if no values are given
    >>> test.invalidate(2)
    >>> test(1)
    2
    >>> test(2)
    called
    3
    """
    method.cache = {}

    def invalidate(*arguments, **keyword_arguments):
        key = _represent_arguments(*arguments, **keyword_arguments)
        if not key:
            method.cache = {}
        elif key in method.cache:
            del method.cache[key]
        else:
            raise KeyError(
                'Not prevously cached: %s(%s)' % (method.__name__, key))

    def new_method(*arguments, **keyword_arguments):
        """Cache the arguments and return values of the call

        The key cached is the repr() of arguments
            This allows more types of values to be used as keys to the cache
            Such as lists and tuples
        """
        key = _represent_arguments(*arguments, **keyword_arguments)
        if key not in method.cache:
            method.cache[key] = method(*arguments, **keyword_arguments)
        return method.cache[key]
    new_method.invalidate = invalidate
    new_method.__doc__ = method.__doc__
    new_method.__name__ = 'memoize(%s)' % method.__name__
    return new_method


def debug(method):
    """Decorator to debug the given method"""
    def new_method(*args, **kwargs):
        import pdb
        try:
            import pudb
        except ImportError:
            pudb = pdb
        try:
            pudb.runcall(method, *args, **kwargs)
        except pdb.bdb.BdbQuit:
            sys.exit('Normal quit from debugger')
    new_method.__doc__ = method.__doc__
    new_method.__name__ = 'debug(%s)' % method.__name__
    return new_method


def argparser(main_method, options, **kwargs):

    def main(argv):
        args = parser.parse_args(argv)
        for method, _flag, name, _help in options:
            if not method:
                continue
            value = getattr(args, name)
            if not value:
                continue
            method(value)
        return main_method(args)

    from argparse import ArgumentParser
    parser = ArgumentParser(description='Process some integers.')
    for _method, flag, name, help in options:
        name_or_flags = [flag, name] if flag else name
        parser.add_argument(name_or_flags, help=help, **kwargs)
    return main


def streamer(main_method):
    """Open a stream for the first file in arguments, or stdin"""
    def main(arguments):
        streams = [StringIO(get_clipboard_data())] if (arguments and '-c' in arguments) else []
        streams = streams or ([file(_, 'r') for _ in arguments if os.path.isfile(argument)] if arguments else [])
        streams = streams or ([sys.stdin] if not (streams or arguments) else [])
        return main_method(arguments, streams)
    return main


def old_streamer(main_method):
    """Open a stream for the first file in arguments, or stdin"""
    if not arguments:
        return [sys.stdin]
    elif arguments[0] == '-c':
        return [StringIO(get_clipboard_data())]
    for argument in arguments:
        if os.path.isfile(argument):
            return file(argument, 'r')
    return method


def globber(main_method, globs):
    """Recognise globs in args"""
    import os
    from glob import glob

    def main(arguments):
        lists_of_paths = [_ for _ in arguments if glob(pathname, recursive=True)]
        return main_method(arguments, lists_of_paths)
    return main



"""Framework for a script to be run from a shell"""

import argparse
import os
import sys

from pysyte import __version__
from pysyte.debuggers import DebugExit


_versions = [__version__]
_exit_ok = os.EX_OK
_exit_fail = not _exit_ok

args = None


def latest_version():
    return _versions[-1]


def version():
    print('%s %s' % (args, latest_version()))
    raise SystemExit


def parse_args(add_args, docstring):
    """Parse out command line arguments"""

    def run_args():
        """Run any global methods eponymous with args"""
        valuable_args = {k for k, v in args.__dict__.items() if v}
        arg_methods = {globals()[a] for a in valuable_args if a in globals()}
        if not arg_methods:
            return
        for method in arg_methods:
            method()
        raise SystemExit(_exit_ok)

    parser = argparse.ArgumentParser(
        description=docstring and docstring.splitlines()[0] or "No docstring")
    result = add_args(parser)
    parser = result if result else parser
    parser.add_argument('-v', '--version', action='store_true',
                        help='Show version [default: %s]' % __version__)
    # parser.add_argument('-q', '--quiet', action='store_true',
    # help='Do not show stdout')
    # parser.add_argument('-Q', '--quiet_errors', action='store_true',
    # help='Do not show stderr')
    global args  # pylint: disable=global-statement
    args = parser.parse_args()
    run_args()
    return args


# pylint: disable=redefined-outer-name
def main(method, add_args, version=None,
         docstring=None, error_stream=sys.stderr):
    """Run the method as a script

    Add a '--version' argument
    Call add_args(parser) to add any more

    Parse sys.argv
    call method(args)
    Catch any exceptions
        exit on SystemExit or KeyboardError
        Others:
            If version is less then 1, reraise them
            Else print them to the error stream
    """
    if version:
        _versions.append(version)
    try:
        args = parse_args(
            add_args,
            docstring and docstring or '%s()' % method.__name__)
        return _exit_ok if method(args) else _exit_fail
    except KeyboardInterrupt as e:
        ctrl_c = 3
        return ctrl_c
    except SystemExit as e:
        return e.code
    except DebugExit:
        return _exit_fail
    except Exception as e:  # pylint: disable=broad-except
        if int(latest_version().split('.')[0]) < 1:
            raise
        if error_stream:
            print(e, file=error_stream)
        return _exit_fail
    return _exit_ok



"""Handle bash commands for the tools package"""

import logging
import os
from contextlib import contextmanager

from pysyte.bash.cmnds import getstatusoutput

logger = logging.getLogger(__name__)


class BashError(ValueError):
    pass


_working_dirs = ['']
_displays = [False]
_path = []


def cd(path):
    if _working_dirs[0] == path:
        return
    logger.info('$ PWD=%s', path)
    _working_dirs[0] = path


@contextmanager
def pushd(path):
    _working_dirs.insert(0, path)
    yield
    del _working_dirs[0]


def _get_path():
    """Guarantee that /usr/local/bin and /usr/bin are in PATH"""
    if _path:
        return _path[0]
    environ_paths = set(os.environ['PATH'].split(':'))
    environ_paths.add('/usr/local/bin')
    environ_paths.add('/usr/bin')
    _path.append(':'.join(environ_paths))
    logger.debug('PATH = %s', _path[-1])
    return _path[0]


def run(command):
    path_command = 'PATH=%s %s' % (_get_path(), command)
    if _working_dirs[0]:
        run_command = '(cd %s; %s)' % (_working_dirs[0], path_command)
        logger.info('$ (cd %s; %s)', _working_dirs[0], command)
    else:
        run_command = path_command
        logger.info('$ %s', command)
    status, output = getstatusoutput(run_command)
    if status:
        logger.error('\n%s', output)
        raise BashError(output)
    elif output:
        logger.info('\n%s', output)
    return output


def set_verbosity(display):
    _displays[0] = display


def full_path(path):
    """Get the real path, expanding links and bashisms"""
    return os.path.realpath(os.path.expanduser(os.path.expandvars(path)))



"""Set up the pysyte project"""

import os
from setuptools import find_packages
from setuptools import setup

import pysyte

description = """Python modules often used in dotjab, and elsewhere

Available on guthub and pypi for ease of access
But probably not of great interest to others
"""

def package_files(directory):
    paths = []
    test_extensions = ('.test', '.tests')
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            extension = os.path.splitext(filename)[-1]
            if extension not in test_extensions:
                continue
            paths.append(os.path.join('..', path, filename))
    return paths

extra_files = package_files('pysyte')


setup(
    name=pysyte.__name__,
    packages=find_packages(),
    package_data={'': package_files('pysyte')},
    version=pysyte.__version__,
    url='https://github.com/jalanb/%s' % pysyte.__name__,
    download_url='https://github.com/jalanb/%s/tarball/v%s' % (
        pysyte.__name__, pysyte.__version__),
    license='MIT License',
    author='J Alan Brogan',
    author_email='github@al-got-rhythm.net',
    description=description.splitlines()[0],
    long_description=description,
    platforms='any',
    classifiers=[
        'Programming Language :: Python :: 3.7',
        'Development Status :: 3 - Alpha',
        'Natural Language :: English',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Build Tools',
    ],
    install_requires=[
        'inflect',
        'pprintpp',
        'path.py==7.7.1',
        'six',
    ],
    scripts=[
        'bin/kat',
        'bin/short_dir',
        'bin/try',
    ],
    test_suite='nose.collector',
    tests_require=['nose'],
    extras_require={
        'testing': ['nose'],
    }
)



"""Module to handle streams of text"""

import os
import sys
import contextlib
from itertools import chain

from six import StringIO

from pysyte import args as arguments
from pysyte import iteration
from pysyte.platforms import get_clipboard_data

def parse_args():
    """Parse out command line arguments"""
    parser = arguments.parser(__doc__)
    parser.args('streams', help='streams to use')
    parser.opt('-p', '--paste', 'paste text from clipboard')
    parser.opt('-i', '--stdin', 'wait for text from stdin')
    return parser.parse_args()


def args(parsed_args, name=None):
    """Interpret parsed args to streams"""
    strings = parsed_args.arg_strings(name)
    files = [s for s in strings if os.path.isfile(s)]
    if files:
        streams = [open(f) for f in files]
    else:
        streams = []
    if getattr(parsed_args, 'paste', not files):
        streams.append(clipboard_stream())
    if getattr(parsed_args, 'stdin', False):
        streams.append(sys.stdin)
    elif not streams:
        streams = [sys.stdin]
    return streams


def clipboard_stream(name=None):
    stream = StringIO(get_clipboard_data())
    stream.name = name or '<clipboard>'
    return stream


def _arg_files():
    return [a for a in sys.argv[1:] if os.path.isfile(a)]


def _arg_streams():
    """yield streams to all arg.isfile()"""
    for path in _arg_files():
        yield open(path)


def all():
    more = iter([clipboard_stream(), sys.stdin])
    return chain(_arg_streams(), more)


def first():
    return iteration.first(all())


def last():
    return iteration.last(all())


def some():
    if sys.argv[1:]:
        assert _arg_files()
    return any()


def _any():
    try:
        stream = iteration.first(_arg_streams())
        if stream:
            return _arg_streams()
    except ValueError:
        return iter([clipboard_stream(), sys.stdin])


def first_file():
    try:
        return iteration.first(_arg_streams())
    except ValueError:
        return None


def full_lines(stream):
    for line in stream.readlines():
        stripped = line.rstrip()
        if stripped:
            yield stripped



try:
    from itertools import ifilter
except ImportError:
    ifilter = filter


def first(sequence, message=None):
    """The first item in that sequence

    If there aren't any, raise a ValueError with that message
    """
    try:
        return next(iter(sequence))
    except StopIteration:
        raise ValueError(message or ('Sequence is empty: %s' % sequence))


def last(sequence, message=None):
    """The last item in that sequence

    If there aren't any, raise a ValueError with that message
    """
    try:
        return sequence.pop()
    except AttributeError:
        return list(sequence).pop()
    except IndexError:
        raise ValueError(message or f'Sequence is empty: {sequence}')


def first_or(sequence, value):
    try:
        return first(sequence)
    except ValueError:
        return value


def first_that(predicate, sequence, message=None):
    """The first item in that sequence that matches that predicate

    If none matches raise a KeyError with that message
    """
    try:
        return next(ifilter(predicate, sequence))
    except StopIteration:
        raise KeyError(message or 'Not Found')
